"""Constants used throughout."""

from collections import OrderedDict

SUPPRESS_LOGS = ["boto3", "botocore", "geopandas", "fiona", "rasterio", "pyogrio", "shapely"]

MAP_DEM_UNCLIPPED_SRC_URL = (
    "https://rockyweb.usgs.gov/vdelivery/Datasets/Staged/Elevation/13/TIFF/USGS_Seamless_DEM_13.vrt"
)

MAP_DEM_BUFFER_DIST_FT = 1000.0
MAP_DEM_DIRNAME = "MapTerrain"
MAP_DEM_HDF_NAME = "MapTerrain.hdf"
MAP_DEM_VERT_UNITS = "Meters"

METERS_PER_FOOT = 1200.0 / 3937.0

MINDEPTH = 0.1  # ft
MIN_FLOW = 1  # cfs

DEFAULT_EPSG = 4269
NORMAL_DEPTH = 0.001

STAC_API_URL = "https://stac2.dewberryanalytics.com"

TERRAIN_NAME = "Terrain_Name"
TERRAIN_PATH = "Terrain_Path"
SUPPORTED_LAYERS = ["River", "XS"]

WSE_HDF_PATH = "/Results/Steady/Output/Output Blocks/Base Output/Steady Profiles/Cross Sections/Water Surface"
FLOW_HDF_PATH = "/Results/Steady/Output/Output Blocks/Base Output/Steady Profiles/Cross Sections/Flow"
XS_NAMES_HDF_PATH = "/Results/Steady/Output/Geometry Info/Cross Section Only"
PROFILE_NAMES_HDF_PATH = "/Results/Steady/Output/Output Blocks/Base Output/Steady Profiles/Profile Names"

LAYER_COLORS = OrderedDict(
    {
        "Banks": "red",
        "Junction": "red",
        "BCLines": "brown",
        "BreakLines": "black",
        "Connections": "cyan",
        "Structure": "black",
        "Mesh": "yellow",
        "River": "blue",
        "StorageAreas": "orange",
        "TwoDAreas": "purple",
        "XS": "green",
    }
)

SHOW_RAS = False

HYDROFABRIC_CRS = 5070



"""Conflate HEC-RAS Model."""

import json
import logging
import os
import traceback
from datetime import datetime
from urllib.parse import quote

import boto3
import pandas as pd
import pystac

import ripple1d
from ripple1d.conflate.plotter import plot_conflation_results
from ripple1d.conflate.rasfim import (
    RasFimConflater,
    nearest_line_to_point,
    ras_reaches_metadata,
    walk_network,
)
from ripple1d.ops.metrics import compute_conflation_metrics
from ripple1d.utils.ripple_utils import clip_ras_centerline

logging.getLogger("fiona").setLevel(logging.ERROR)
logging.getLogger("botocore").setLevel(logging.ERROR)


def href_to_vsis(href: str, bucket: str) -> str:
    """Convert public aws href to a virtual ref for gdal read."""
    return href.replace(f"https://{bucket}.s3.amazonaws.com", f"/vsis3/{bucket}")


def s3_public_href(bucket: str, key: str) -> str:
    """Convert bucket and key to public href."""
    return f"https://{bucket}.s3.amazonaws.com/{quote(key)}"


def conflate_single_nwm_reach(rfc: RasFimConflater, nwm_reach_id: int):
    """Conflate a HEC-RAS model with a specific NWM reach."""
    nwm_reach_id_identified = False
    model_local_nwm_reaches = rfc.local_nwm_reaches()
    for river_reach_name in rfc.ras_river_reach_names:
        local_nwm_reaches = model_local_nwm_reaches.intersects(
            rfc.ras_xs_concave_hull(river_reach_name)["geometry"].iloc[0]
        )

        ras_start_point, ras_stop_point = rfc.ras_start_end_points(river_reach_name=river_reach_name)
        us_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_start_point)
        ds_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_stop_point)

        potential_reach_path = walk_network(local_nwm_reaches, us_most_reach_id, ds_most_reach_id)
        candidate_reaches = local_nwm_reaches.query(f"ID in {potential_reach_path}")
        if len(candidate_reaches.query(f"ID == {nwm_reach_id}")) == 1:
            nwm_reach_id_identified = True
            return ras_reaches_metadata(rfc, candidate_reaches[candidate_reaches["ID"] == nwm_reach_id])
    if not nwm_reach_id_identified:
        raise ValueError(f"nwm_reach_id {nwm_reach_id} not conflating to the ras model geometry.")


def conflate_model(source_model_directory: str, source_network: dict):
    """Conflate a HEC-RAS model with NWM reaches.

    Parameters
    ----------
    source_model_directory : str
        The path to the directory containing HEC-RAS project, plan, geometry,
        and flow files.
    source_network : dict
        Information on the network to conflate

        - **file_name** (str):
            path/to/nwm_network.parquet (required)
        - **type** (str):
            must be 'nwm_hydrofabric' (required)
        - **version** (str):
            optional version number to log
    task_id : str, optional
        Task ID to use for logging, by default ""

    Returns
    -------
    str
        Path to the .conflation.json file generated

    Raises
    ------
    KeyError
        Raises when source_network dict does not contain a file_name value
    ValueError
        Raises when source_network type is not nwm_hydrofabric

    Notes
    -----
    The spatial extents of HEC-RAS river reaches and National Water model (NWM)
    reaches are not aligned.  The conflate_model endpoint resolves these
    differences by associating HEC-RAS models and model components (e.g.
    cross-sections) with the NWM reaches they overlap.

    #. Generate a concave hull (bounding geometry) around the HEC-RAS source
       model cross-sections
    #. Extract NWM reaches intersecting the hull
    #. For each HEC-RAS river reach within the source model,
      #. Locate the NWM reaches nearest to the most upstream and most
         downstream cross-sections
      #. Extract all intermediate NWM reaches by walking   the network from
         upstream to downstream
    #. For each NWM reach extracted,
      #. Locate the HEC-RAS cross-sections that intersect the reach
        #. Discard cross-sections that are not drawn right to left looking
           downstream
        #. If no cross-sections intersect the reach, mark the reach as
           “eclipsed”
      #. Mark the HEC-RAS cross-section closest to the upstream end of the
         reach as the “us_xs ”
      #. Identify the HEC-RAS cross-section that is closest to the downstream
         end of the reach, and then mark the next HEC-RAS cross-section
         downstream of it as the “ds_xs”
        #. If the “ds_xs” would be a HEC-RAS junction, mark the first
           cross-section downstream of the junction as “ds_xs”
      #. If “ds_xs” and “us_xs” are the same, mark the reach as eclipsed
    #. Generate a map of the conflated reaches and calculate conflation metrics

    Additionally, high and low flows are generated for each reach to bound the
    SRC generated in later steps. The low flow is 1.2 times the high flow
    threshold listed for the reach in the NWM network.  The high flow is the
    100 year flow from the NWM network.
    """
    logging.info(f"conflate_model starting")
    try:
        nwm_pq_path = source_network["file_name"]
    except KeyError:
        raise KeyError(f"source_network must contain 'file_name', invalid parameters: {source_network}")

    if not source_network["type"] == "nwm_hydrofabric":
        raise ValueError(f"source_network type must be 'nwm_hydrofabric', invalid parameters: {source_network}")

    version = source_network.get("version", "")

    rfc = RasFimConflater(nwm_pq_path, source_model_directory)
    metadata = {"reaches": {}}
    buffer = 1000
    for river_reach_name in rfc.ras_river_reach_names:
        try:
            local_nwm_reaches = rfc.local_nwm_reaches(river_reach_name, buffer=buffer)

            # get the start and end points of the river reach
            ras_start_point, ras_stop_point = rfc.ras_start_end_points(
                river_reach_name=river_reach_name, clip_to_xs=True
            )

            # if this is the only river reach, raise error
            river = rfc.ras_centerline_by_river_reach_name(river_reach_name)
            river = clip_ras_centerline(river, rfc.xs_by_river_reach_name(river_reach_name))
            truncate_distance = 0
            us_most_reach_id, ds_most_reach_id = None, None
            while True:
                try:
                    us_most_reach_id = nearest_line_to_point(
                        local_nwm_reaches, ras_start_point, start_reach_distance=100
                    )
                    break
                except ValueError as e:
                    if truncate_distance > 10000:
                        logging.info(f"Could not identifiy a network reach near the upstream end of {river_reach_name}")
                        break
                    truncate_distance += 100
                    ras_start_point = river.interpolate(truncate_distance)

                    logging.debug(
                        f"truncate_distance: {truncate_distance} | length {ras_start_point} |  river reach: {river_reach_name}"
                    )
            if us_most_reach_id is None:
                continue

            try:
                ds_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_stop_point)
            except:
                logging.info(f"Could not identifiy a network reach near the downstream end of {river_reach_name}")
                continue

            logging.info(
                f"{river_reach_name} | us_most_reach_id ={us_most_reach_id} and ds_most_reach_id = {ds_most_reach_id}"
            )

            # walk network to get the potential reach ids
            potential_reach_path = walk_network(local_nwm_reaches, us_most_reach_id, ds_most_reach_id, river_reach_name)
            potential_reach_path = list(set(potential_reach_path) - set(metadata.keys()))

            # get gdf of the candidate reaches
            candidate_reaches = local_nwm_reaches.query(f"ID in {potential_reach_path}")

            metadata["reaches"].update(ras_reaches_metadata(rfc, candidate_reaches, river_reach_name))
        except Exception as e:
            logging.error(f"river-reach: {river_reach_name} | Error: {e}")
            logging.error(f"river-reach: {river_reach_name} | Traceback: {traceback.format_exc()}")

    # if not conflated(metadata):
    #     return f"no reaches conflated"

    ids = list(metadata["reaches"].keys())
    fim_stream = rfc.local_nwm_reaches()[rfc.local_nwm_reaches()["ID"].isin(ids)]
    # conflation_png = f"{rfc.ras_gpkg.replace('.gpkg','.conflation.png')}"

    # plot_conflation_results(
    #     rfc,
    #     fim_stream,
    #     conflation_png,
    #     limit_plot_to_nearby_reaches=True,
    # )

    logging.debug(f"Conflation results: {metadata}")
    conflation_file = f"{rfc.ras_gpkg.replace('.gpkg','.conflation.json')}"

    metadata["metadata"] = {}
    metadata["metadata"]["source_network"] = source_network.copy()
    metadata["metadata"]["source_network"]["file_name"] = os.path.basename(nwm_pq_path)
    # metadata["metadata"]["conflation_png"] = os.path.basename(conflation_png)
    metadata["metadata"]["conflation_ripple1d_version"] = ripple1d.__version__
    metadata["metadata"]["metrics_ripple1d_version"] = ripple1d.__version__
    metadata["metadata"]["source_ras_model"] = {
        "stac_api": rfc.stac_api,
        "stac_collection_id": rfc.stac_collection_id,
        "stac_item_id": rfc.stac_item_id,
    }
    metadata["metadata"]["source_ras_model"]["source_ras_files"] = {
        "geometry": rfc.primary_geom_file,
        "forcing": rfc.primary_flow_file,
        "project-file": rfc.ras_project_file,
        "plan": rfc.primary_plan_file,
    }

    with open(conflation_file, "w") as f:
        f.write(json.dumps(metadata, indent=4))

    try:
        compute_conflation_metrics(source_model_directory, source_network)
    except Exception as e:
        logging.error(f"| Error: {e}")
        logging.error(f"| Traceback: {traceback.format_exc()}")

    logging.info(f"conflate_model complete")
    return {"conflation_file": conflation_file}


def conflated(metadata: dict) -> bool:
    """Determine if any reaches conflated."""
    count = 0
    for reach_data in metadata["reaches"].values():
        if not reach_data["eclipsed"]:
            count += 1
    if count == 0:
        return False
    else:
        return True



"""RASMap Contents."""

from .consts import TERRAIN_NAME, TERRAIN_PATH

RASMAP_631 = r"""<RASMapper>
  <Version>2.0.0</Version>
  <Features>
    <Layer Name="Profile Lines" Type="PolylineFeatureLayer" Filename=".\Features\Profile Lines.shp">
      <LabelFeatures Checked="True" PercentPosition="0" rows="1" cols="1" r0c0="Name" Position="3" Color="-16777216" FontSize="8.25" />
    </Layer>
  </Features>
  <Geometries>
    <Layer Name="geom_name_placeholder" Type="RASGeometry" Expanded="True" Filename=".\geom_hdf_placeholder">
      <Layer Type="RASXS" UnitsRiverStation="Feet" RiverStationDecimalPlaces="0" />
      <Layer Type="FinalNValueLayer">
        <ResampleMethod>near</ResampleMethod>
        <Surface On="True" />
      </Layer>
      <Layer Name="TEMP N-Value Rasterizer" Type="InterpretationRasterizerLayer">
        <ResampleMethod>near</ResampleMethod>
        <Surface On="True" />
      </Layer>
      <Layer Name="Final Values" Type="InterpretationRasterizerLayer">
        <ResampleMethod>near</ResampleMethod>
        <Surface On="True" />
      </Layer>
      <Layer Name="Final Values" Type="InterpretationRasterizerLayer">
        <ResampleMethod>near</ResampleMethod>
        <Surface On="True" />
      </Layer>
    </Layer>
  </Geometries>
  <EventConditions />
  <Results />
  <MapLayers />
  <Terrains />
  <CurrentView>
    <MaxX>3763908.23</MaxX>
    <MinX>3747983.37</MinX>
    <MaxY>10229616.82</MaxY>
    <MinY>10223235.59</MinY>
  </CurrentView>
  <VelocitySettings>
    <Density>1.5</Density>
    <Lifetime>100</Lifetime>
    <Radius>0.75</Radius>
    <Method>2</Method>
    <Timestep>1</Timestep>
    <StaticColor>Black</StaticColor>
    <SpeedRelativeToZoom>False</SpeedRelativeToZoom>
  </VelocitySettings>
  <AnimationSettings>
    <DelayTimer>0</DelayTimer>
  </AnimationSettings>
  <Units>US Customary</Units>
  <RenderMode>sloping</RenderMode>
  <ReduceShallowToHorizontal>true</ReduceShallowToHorizontal>
  <MarksWarpMethod>False</MarksWarpMethod>
  <CurrentSettings>
    <ProjectSettings>
      <RiverStationUnits>Feet</RiverStationUnits>
      <RiverStationDecimalPlaces>0</RiverStationDecimalPlaces>
      <HorizontalDecimalPlaces>1</HorizontalDecimalPlaces>
      <VerticalDecimalPlaces>2</VerticalDecimalPlaces>
      <XSMaxPoints>450</XSMaxPoints>
      <LSMaxPoints>1000</LSMaxPoints>
      <ProfilePointMinCount>0</ProfilePointMinCount>
      <ShowLegend>True</ShowLegend>
      <ShowNorthArrow>False</ShowNorthArrow>
      <ShowScaleBar>True</ShowScaleBar>
      <ShowGreaterThanInLegend>False</ShowGreaterThanInLegend>
      <MeshTolerance_MinFaceLength>0.05</MeshTolerance_MinFaceLength>
      <MeshTolerance_EnsureCellPoint>False</MeshTolerance_EnsureCellPoint>
    </ProjectSettings>
    <Folders />
  </CurrentSettings>
</RASMapper>"""

TERRAIN = rf"""  <Terrains Checked="True" Expanded="True">
    <Layer Name="{TERRAIN_NAME}" Type="TerrainLayer" Checked="True" Filename=".\{TERRAIN_PATH}">
      <Symbology>
        <SurfaceFill Colors="-10039894,-256,-16744448,-23296,-7667712,-5952982,-8355712,-1286" Values="-0.625,50.4063185155618,70.7601808393692,101.843506660136,135.379737550828,199.971461686389,225.529762277054,366.78125" Stretched="True" AlphaTag="255" UseDatasetMinMax="False" RegenerateForScreen="False" />
      </Symbology>
      <ResampleMethod>near</ResampleMethod>
      <Surface On="True" />
    </Layer>
  </Terrains>"""

PLAN = r"""  <Results Expanded="True">
    <Layer Name="plan_name_placeholder" Type="RASResults" Expanded="True" Filename=".\plan_hdf_placeholder">
      <Layer Name="Event Conditions" Type="RASEventConditions" Filename=".\plan_hdf_placeholder">
        <Layer Name="Wind Layer" Type="ResultWindLayer" Filename=".\plan_hdf_placeholder">
          <ResampleMethod>near</ResampleMethod>
          <Surface On="True" />
          <Metadata BandIndex="0" SubDataset="" />
        </Layer>
      </Layer>
      <Layer Type="RASGeometry" Filename=".\plan_hdf_placeholder">
        <Layer Type="RASXS" UnitsRiverStation="Feet" RiverStationDecimalPlaces="0" />
        <Layer Type="RASD2FlowArea">
          <DataColumnGroupVisibilities>
            <DataColumnGroupVisibility DisplayName="Feature Parameters" IsVisible="False" />
          </DataColumnGroupVisibilities>
        </Layer>
        <Layer Type="MeshPerimeterLayer">
          <DataColumnGroupVisibilities>
            <DataColumnGroupVisibility DisplayName="Feature Parameters" IsVisible="True" />
          </DataColumnGroupVisibilities>
        </Layer>
        <Layer Name="Culvert Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Culvert Barrels" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Bridges/Culverts" />
        <Layer Name="Gate Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Gate Openings" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Inline Structures" />
        <Layer Name="Culvert Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Culvert Barrels" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Inline Structures" />
        <Layer Name="Rating Curve Outlets" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Inline Structures" />
        <Layer Name="Outlet Time Series" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Inline Structures" />
        <Layer Name="Gate Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Gate Openings" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Lateral Structures" />
        <Layer Name="Culvert Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Culvert Barrels" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Lateral Structures" />
        <Layer Name="Rating Curve Outlets" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Lateral Structures" />
        <Layer Name="Outlet Time Series" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="Lateral Structures" />
        <Layer Name="Gate Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Gate Openings" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="SA/2D Connections" />
        <Layer Name="Culvert Groups" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="" />
        <Layer Name="Culvert Barrels" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="SA/2D Connections" />
        <Layer Name="Rating Curve Outlets" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="SA/2D Connections" />
        <Layer Name="Outlet Time Series" Type="VirtualGeometryFeatureLayer" ParentIdentifiers="SA/2D Connections" />
        <Layer Type="StructureLayer" UnitsRiverStation="Feet" RiverStationDecimalPlaces="0" />
        <Layer Type="FinalNValueLayer">
          <ResampleMethod>near</ResampleMethod>
          <Surface On="True" />
        </Layer>
        <Layer Name="TEMP N-Value Rasterizer" Type="InterpretationRasterizerLayer">
          <ResampleMethod>near</ResampleMethod>
          <Surface On="True" />
        </Layer>
        <Layer Name="Final Values" Type="InterpretationRasterizerLayer">
          <ResampleMethod>near</ResampleMethod>
          <Surface On="True" />
        </Layer>
        <Layer Name="Final Values" Type="InterpretationRasterizerLayer">
          <ResampleMethod>near</ResampleMethod>
          <Surface On="True" />
        </Layer>
      </Layer>
      <Layer Name="Depth" Type="RASResultsMap">
        <MapParameters MapType="depth" ProfileIndex="0" ProfileName="profile_placeholder" />
      </Layer>
      <Layer Name="Velocity" Type="RASResultsMap">
        <MapParameters MapType="velocity" ProfileIndex="0" ProfileName="profile_placeholder" />
      </Layer>
      <Layer Name="WSE" Type="RASResultsMap">
        <MapParameters MapType="elevation" ProfileIndex="0" ProfileName="profile_placeholder" />
      </Layer>
  </Results>"""



"""Endpoint utilites for executing ripple processes using subprocess calls."""

import json
import logging
import subprocess
import sys
from typing import Callable

from ripple1d.consts import SUPPRESS_LOGS
from ripple1d.ops.fim_lib import create_fim_lib, create_rating_curves_db, fim_lib_stac, nwm_reach_model_stac
from ripple1d.ops.metrics import compute_conflation_metrics
from ripple1d.ops.ras_conflate import conflate_model
from ripple1d.ops.ras_run import (
    create_model_run_normal_depth,
    run_incremental_normal_depth,
    run_known_wse,
)
from ripple1d.ops.ras_terrain import create_ras_terrain
from ripple1d.ops.subset_gpkg import extract_submodel
from ripple1d.ras_to_gpkg import gpkg_from_ras
from ripple1d.ripple1d_logger import RippleLogFormatter  # , initialize_process_logger

func_lookup = {
    "extract_submodel": extract_submodel,
    "conflate_model": conflate_model,
    "compute_conflation_metrics": compute_conflation_metrics,
    "gpkg_from_ras": gpkg_from_ras,
    "create_ras_terrain": create_ras_terrain,
    "create_model_run_normal_depth": create_model_run_normal_depth,
    "run_incremental_normal_depth": run_incremental_normal_depth,
    "run_known_wse": run_known_wse,
    "create_fim_lib": create_fim_lib,
    "nwm_reach_model_stac": nwm_reach_model_stac,
    "fim_lib_stac": fim_lib_stac,
    "create_rating_curves_db": create_rating_curves_db,
}


def execute_endpoint(func: str, args: dict):
    """Execute the specified function."""
    return func_lookup[func](**json.loads(args))


def initialize_process_logger(level, verbose: bool = False):
    """Configure logging for ripple1d."""
    datefmt = "%Y-%m-%dT%H:%M:%SZ"

    for module in SUPPRESS_LOGS:
        logging.getLogger(module).setLevel(logging.WARNING)

    handler = logging.StreamHandler(sys.stdout)
    logging.basicConfig(
        level=level,
        handlers=[handler],
        format="""{"time": "%(asctime)s" , "level": "%(levelname)s", "msg": "%(message)s"}""",
    )


if __name__ == "__main__":
    initialize_process_logger(logging.INFO)
    try:
        r = execute_endpoint(sys.argv[1], sys.argv[2])
        print((json.dumps({"results": r})))
    except Exception as e:
        raise SystemError(f"Exception occurred: {e}")
        sys.exit(1)



"""huey instance for huey + Flask REST API (to be called by huey_consumer.py)."""

import ast
import json
import logging
import os
import subprocess
import sys
import time
import typing
from math import exp

import psutil
from huey import SqliteHuey, signals
from huey.api import Result
from matplotlib.image import resample

from ripple1d.ripple1d_logger import initialize_server_logger

initialize_server_logger()

huey = SqliteHuey(filename=os.path.join("jobs.db"), results=True)

# Create custom table task_status
huey.storage.sql(
    """
    create table if not exists "task_status" (
        "task_id" text not null,
        "p_id" text,
        "func_name" text not null,
        "func_kwargs" text not null,
        "huey_status" text not null,
        "ogc_status" text not null,
        "accept_time" timestamptz not null,
        "dismiss_time" timestamptz,
        "start_time" timestamptz,
        "finish_time" timestamptz,
        "status_time" timestamptz,
        "finish_duration_minutes" real,
        "results" text,
        primary key("task_id")
    );
    """,
    commit=True,
)

# Create custom table task_logs
huey.storage.sql(
    """
    create table if not exists "task_logs" (
        "task_id" text not null,
        "stdout" text,
        "stderr" text,
        "results" text,
        primary key("task_id")
    );
    """,
    commit=True,
)

# Create custom triggers to keep status_time field updated
huey.storage.sql(
    """
    create trigger if not exists "trg_task_status_insert" after insert on "task_status"
    begin
        update "task_status" set
            "status_time" = max(
            ifnull("accept_time", datetime(0)),
            ifnull("dismiss_time", datetime(0)),
            ifnull("start_time", datetime(0)),
            ifnull("finish_time", datetime(0))
        ),
            "finish_duration_minutes" = 1440.0 * (julianday("finish_time") - julianday("start_time"))
        where "task_id" = NEW."task_id";
    end;
    """,
    commit=True,
)

huey.storage.sql(
    """
    create trigger if not exists "trg_task_status_update" after update on "task_status"
    begin
        update "task_status" set
            "status_time" = max(
            ifnull("accept_time", datetime(0)),
            ifnull("dismiss_time", datetime(0)),
            ifnull("start_time", datetime(0)),
            ifnull("finish_time", datetime(0))
        ),
            "finish_duration_minutes" = 1440.0 * (julianday("finish_time") - julianday("start_time"))
        where "task_id" = NEW."task_id";
    end;
    """,
    commit=True,
)


def create_and_enqueue_task(func: typing.Callable, kwargs: dict = {}, priority: int = None) -> Result:
    """Create a task instance, then add it to the custom table "task_status" before adding it to the queue.

    Return the Result object associated with the task, which is the same class type that would be returned by
    calling the @huey.task()-decorated function itself (in this case, `_process`).

    This is needed to avoid a race condition, this ensures that the task exists in the custom table before
    it starts executing.
    """
    task_instance = _process.s(func, kwargs, priority=priority)
    huey.storage.sql(
        """
        insert into "task_status"
            ("task_id", "func_name", "func_kwargs", "huey_status", "ogc_status", "accept_time")
        values
            (?, ?, ?, 'queued', 'accepted', datetime('now'))""",
        (task_instance.id, func.__name__, json.dumps(kwargs)),
        True,
    )
    return huey.enqueue(task_instance)


def revoke_task_by_pid(task_id: str):
    """Revoke a task by pid."""
    expression = """select "p_id" from "task_status" where "task_id" = ?"""
    args = (task_id,)
    pid = huey.storage.sql(expression, args, results=True)[0][0]
    psutil.Process(int(pid)).terminate()
    expression = f"""
        update "task_status"
        set
            "huey_status" = 'complete',
            "ogc_status" = 'dismissed',
            "dismiss_time" = datetime('now')
        where "task_id" = ?
        """
    args = (task_id,)
    huey.storage.sql(expression, args, True)


def revoke_task(task_id: str):
    """Revoke a task."""
    huey.revoke_by_id(task_id)
    expression = f"""
        update "task_status"
        set
            "ogc_status" = 'dismissed',
            "dismiss_time" = datetime('now')
        where "task_id" = ?
        """
    args = (task_id,)
    huey.storage.sql(expression, args, True)


def update_p_id(task_id: str, p_id: str):
    """Update p_id."""
    expression = f"""
        update "task_status"
        set
            "p_id" = '{p_id}'
        where "task_id" = ?
        """
    args = (task_id,)
    huey.storage.sql(expression, args, True)

def task_status(only_task_id: str | None) -> dict[str, dict]:
    """Return dictionary of tasks, where key is task ID and value is subdict (status fields related to each task).

    If only_task_id is None, all tasks will be returned.
    If only_task_id is *not* None, only the provided task will be returned.
    """
    fields = ["task_id","func_name","func_kwargs","huey_status","ogc_status","accept_time","dismiss_time","start_time","finish_time","status_time","finish_duration_minutes"]
    if only_task_id is not None:
        filterer = f' WHERE task_id = "{only_task_id}"'
    else:
        filterer = ''
    expression = f'SELECT {", ".join(fields)} FROM task_status{filterer} ORDER BY status_time DESC;'
    results_raw = huey.storage.sql(expression, results=True)
    results_dict = {}
    for r in results_raw:
        task_id = r[0]
        results_dict[task_id] = {k: v for k, v in zip(fields[1:], r[1:])}
    return results_dict

def task_results(only_task_id: str | None) -> dict[str, dict]:
    """Return dictionary of tasks, where key is task ID and value is subdict with val, err, and tb.

    If only_task_id is None, all tasks will be returned.
    If only_task_id is *not* None, only the provided task will be returned.
    """
    fields = ["task_id","stderr","results"]
    if only_task_id is None:
        only_task_id = '%'
    expression = f'SELECT {", ".join(fields)} FROM task_logs WHERE task_id = "{only_task_id}";'
    results_raw = huey.storage.sql(expression, results=True)
    results_dict = {}
    for r in results_raw:
        task_id = r[0]
        results_dict[task_id] = {}
        results_dict[task_id]['val'] = r[2]
        if r[1] is not None:
            results_dict[task_id]['err'] = r[1].splitlines()[-1]
        else:
            results_dict[task_id]['err'] = None
        results_dict[task_id]['tb'] = r[1]
    return results_dict

def task_summary(only_task_id: str | None) -> dict[str, dict]:
    """Return dictionary of tasks, where key is task ID and value is subdict with status and results.

    If only_task_id is None, all tasks will be returned.
    If only_task_id is *not* None, only the provided task will be returned.
    """
    status = task_status(only_task_id)
    results = task_results(only_task_id)
    for t in status:
        if t in results:
            status[t]['result'] = results[t]
        else:
            status[t]['result'] = None
    return status

def fetch_one_query(task_id: str, field: str, table: str) -> str:
    """Return a fetch one query given table and field given task_id."""
    expression = f"""select "{field}" from "{table}" where "task_id" = ?"""
    results = huey.storage.sql(expression, (task_id,), results=True)
    if not results:
        return "notfound"
    return results[0][0]


def huey_status(task_id: str) -> str:
    """For given task ID, return its current status in huey terms."""
    # expression = """select "huey_status" from "task_status" where "task_id" = ?"""
    # results = huey.storage.sql(expression, (task_id,), results=True)
    # if not results:
    #     return "notfound"
    # return results[0][0]
    return fetch_one_query(task_id, "huey_status", "task_status")


def fetch_ogc_status(task_id: str) -> str:
    """For given task ID, return its current status in OGC API terms."""
    return fetch_one_query(task_id, "ogc_status", "task_status")


def fetch_results(task_id: str) -> str:
    """For given task ID, return results."""
    results = fetch_one_query(task_id, "results", "task_logs")
    return json.loads(results)


def fetch_logs(task_id: str) -> str:
    """For given task ID, return logs."""
    expression = f"""
        select
            "stdout",
            "stderr"
        from "task_logs"
        where "task_id" = ?"""

    results_raw = huey.storage.sql(expression, (task_id,), results=True)

    results = {"logs": [], "errors": []}
    try:
        std_out = ast.literal_eval(results_raw[0][0])
        for line in std_out:
            results["logs"].append(json.loads(line))
    except:
        results["logs"] = results_raw[0][0]

    try:
        std_err = ast.literal_eval(results_raw[0][1])
        for line in std_err:
            results["errors"].append(json.loads(line))
        else:
            results["errors"].append(results_raw[0][1])
    except:
        results["errors"] = results_raw[0][1]

    return results


def job_dismissed(task_id: str) -> bool:
    """For given task ID, return its current status in OGC API terms."""
    expression = """select "dismiss_time" from "task_status" where "task_id" = ?"""
    results = huey.storage.sql(expression, (task_id,), results=True)
    logging.debug(f"job_dismissed response = {results}")
    if results[0][0] == "None" or results[0][0] is None:
        return False
    return True

def noop(task_id: str = None):
    """Do nothing except log a message. For ping endpoint and testing."""
    logging.info(f"{task_id} | noop")
    pass


def subprocess_caller(
    func: str, args: dict, task_id: str, log_dir: str = "", log_level: int = logging.INFO, timeout: int = None
):
    """Call the specified function through a subprocess."""
    subprocess_args = [
        sys.executable,
        "-u",
        os.path.dirname(__file__).replace("api", "ops/endpoints.py"),
        func,
        json.dumps(args),
    ]

    process = subprocess.Popen(
        subprocess_args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    results, pid = None, None
    update_p_id(task_id, process.pid)

    try:
        stdout_lines, stderr_output = process.communicate(timeout=timeout)
        logs = stdout_lines.splitlines()
        for i, line in enumerate(logs):
            if '"results"' in line:
                program_output = json.loads(line)
                results = json.dumps(program_output.get("results", None))
                del logs[i]
                pid = program_output.get("results", None).get("pid", None)

    except subprocess.TimeoutExpired:
        process.kill()
        stderr_output.append("Process timed out")
        raise TimeoutError(f"{task_id}: Process timed out")

    except Exception as e:
        stderr_output.append(str(e))
        raise SystemError(f"{task_id}: Exception occurred: {e}")

    exit_code = process.wait()

    if len(stderr_output) == 0:
        errors = None
    else:
        errors = str(stderr_output)

    # for RAS computes only
    if pid is not None:
        # update database pid to be ras pid
        huey.storage.sql("""update "task_status" set "p_id" = ? where "task_id" = ?""", (pid, task_id), True)
        # wait for pid status to be finished.
        while psutil.pid_exists(pid):
            time.sleep(1)

    huey.storage.sql(
        """
        insert into "task_logs"
            ("task_id", "stdout", "stderr", "results")
        values
            (?, ?, ?, ?)""",
        (task_id, str(logs), errors, results),
        True,
    )

    if exit_code == 15:
        huey.storage.sql(
            """update "task_status" set "ogc_status" = ? where "task_id" = ?""", ("dismissed", task_id), True
        )

    elif exit_code != 0:
        logging.debug(f"{task_id} exit code {exit_code}")
        huey.storage.sql("""update "task_status" set "ogc_status" = ? where "task_id" = ?""", ("failed", task_id), True)


@huey.task(context=True)
def _process(func: typing.Callable, kwargs: dict = {}, task=None):
    """Execute generic huey task that calls the provided func with provided kwargs, asynchronously."""
    if task:
        task_id = task.id
    else:
        task_id = None
    return subprocess_caller(func.__name__, kwargs, task_id)


@huey.signal()
def _handle_signals(signal, task, exc=None):
    """Update the status in the task_status table When task emits a signal."""
    # logging.info(f"{signal} : {task.id}")
    task_status = signal
    match signal:
        case signals.SIGNAL_EXECUTING:
            time_field = "start_time"
            ogc_status = "running"

        case signals.SIGNAL_COMPLETE:
            time_field = "finish_time"
            if job_dismissed(task.id):
                task_status = "revoked"
                ogc_status = "dismissed"
                huey.storage.sql(
                    """update "task_status" set huey_status = ? where "task_id" = ?""", (task_status, task.id), True
                )
            elif fetch_ogc_status(task.id) == "failed":
                task_status = "complete"
                ogc_status = "failed"
            else:
                ogc_status = "successful"

        case signals.SIGNAL_ERROR:
            time_field = "finish_time"
            ogc_status = "failed"

        case signals.SIGNAL_LOCKED | signals.SIGNAL_EXPIRED:
            time_field = "finish_time"
            ogc_status = "failed"

        case signals.SIGNAL_CANCELED | signals.SIGNAL_INTERRUPTED:
            time_field = "dismiss_time"
            ogc_status = "dismissed"

        case signals.SIGNAL_REVOKED:
            # Set huey status, then short-circuit without setting "ogc_status" or time field.
            # OGC status and dismiss time field are handled at time of revoke call, rather than waiting
            # for the revoke signal which happens later.
            huey.storage.sql(
                """update "task_status" set huey_status = ? where "task_id" = ?""", (signal, task.id), True
            )
            ogc_status = "dismissed"

        case _:  # e.g. SIGNAL_RETRYING, SIGNAL_SCHEDULED
            raise ValueError(f"Unhandled signal: {signal}")

    if ogc_status == "dismissed":
        return

    expression = f"""
        update "task_status"
        set
            "huey_status" = ?,
            "ogc_status" = ?,
            "{time_field}" = datetime('now')
        where "task_id" = ?
        """

    args = (task_status, ogc_status, task.id)
    huey.storage.sql(expression, args, True)



# """API Lofging module for Ripple1D."""

# import inspect
# import json
# import logging
# import os
# import time
# import traceback
# from datetime import datetime, timezone

# from ripple1d.consts import SUPPRESS_LOGS
# from ripple1d.ripple1d_logger import RippleLogFormatter


# def initialize_log(log_dir: str = "", log_level: int = logging.INFO) -> logging.Logger:
#     """Initialize log with JSON-LD style formatting and throttled level for AWS libs.

#     By default sends to StreamHandler (stdout/stderr), but can provide a filename to log to disk instead.
#     """
#     filename = os.path.join(log_dir, f"{_get_log_filename_suffix()}.jsonld")

#     for module in SUPPRESS_LOGS:
#         logging.getLogger(module).setLevel(logging.ERROR)

#     log = logging.getLogger()
#     log.setLevel(log_level)
#     formatter = RippleLogFormatter()

#     if filename:
#         print(f"Initializing log file: {filename}")
#         os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
#         file_handler = logging.FileHandler(filename=filename)
#         file_handler.setFormatter(formatter)
#         log.addHandler(file_handler)
#     else:
#         stream_handler = logging.StreamHandler()
#         stream_handler.setFormatter(formatter)
#         log.addHandler(stream_handler)

#     return log


# def _get_log_filename_suffix():
#     stack_filenames = [frame.filename for frame in inspect.stack()]

#     # Check for specific patterns in the stack filenames
#     if any("huey" in filename for filename in stack_filenames):
#         return "huey"
#     elif any("flask" in filename for filename in stack_filenames):
#         return "flask"
#     else:
#         raise ValueError(
#             f"Could not determine if process invoked by huey or by flask. Stack filenames: {stack_filenames}"
#         )



"""Create HEC-RAS Terrains."""

from __future__ import annotations

import json
import logging
import os
from pathlib import Path

import geopandas as gpd
import rasterio
from pyproj import CRS

from ripple1d.consts import (
    MAP_DEM_BUFFER_DIST_FT,
    MAP_DEM_UNCLIPPED_SRC_URL,
    MAP_DEM_VERT_UNITS,
    METERS_PER_FOOT,
)
from ripple1d.data_model import NwmReachModel
from ripple1d.ras import create_terrain
from ripple1d.utils.dg_utils import clip_raster, reproject_raster
from ripple1d.utils.ripple_utils import fix_reversed_xs, xs_concave_hull


def get_geometry_mask(gdf_xs_conc_hull: str, MAP_DEM_UNCLIPPED_SRC_URL: str) -> gpd.GeoDataFrame:
    """Get a geometry mask for the DEM based on the cross sections."""
    # build a DEM mask polygon based on the XS extents

    # Buffer the concave hull by transforming it to Albers, buffering it, then transforming it to the src raster crs
    with rasterio.open(MAP_DEM_UNCLIPPED_SRC_URL) as src:
        gdf_xs_conc_hull_buffered = (
            gdf_xs_conc_hull.to_crs(epsg=5070).buffer(MAP_DEM_BUFFER_DIST_FT * METERS_PER_FOOT).to_crs(src.crs)
        )

    if len(gdf_xs_conc_hull_buffered) != 1:
        raise ValueError(f"Expected 1 record in gdf_xs_conc_hull_buffered, got {len(gdf_xs_conc_hull_buffered)}")
    return gdf_xs_conc_hull_buffered.iloc[0]


def write_projection_file(crs: CRS, terrain_directory: str) -> str:
    """Write a projection file for the terrain."""
    projection_file = os.path.join(terrain_directory, "projection.prj")
    with open(projection_file, "w") as f:
        f.write(CRS(crs).to_wkt("WKT1_ESRI"))
    return projection_file


def create_ras_terrain(
    submodel_directory: str,
    terrain_source_url: str = MAP_DEM_UNCLIPPED_SRC_URL,
    vertical_units: str = MAP_DEM_VERT_UNITS,
    resolution: float = None,
    resolution_units: str = None,
):
    """Create a RAS terrain file.

    Parameters
    ----------
    submodel_directory : str
        The path to the directory containing a submodel geopackage
    terrain_source_url : str, optional
        URL to a vrt or other raster with topographic data, by default
        MAP_DEM_UNCLIPPED_SRC_URL
    vertical_units : str, optional
        label for the vertical units of the source terrain (ex. Meters), by
        default MAP_DEM_VERT_UNITS
    resolution : float, optional
        horizontal resolution that terrain will be projected to, by default
        None
    resolution_units : str, optional
        unit for resolution parameter, by default None
    task_id : str, optional
        Task ID to use for logging, by default ""

    Returns
    -------
    str
        descriptive string detailing where terrain was saved

    Raises
    ------
    ValueError
        Raised when resolution arg has been provided but resolution_units arg
        has not been provided
    ValueError
        Raised when resolution_units are not Feet or Meters
    FileNotFoundError
        Raised when no geopackage file is found in submodel_directory

    Notes
    -----
    The create_ras_terrain endpoint downloads a digital elevation model (DEM)
    for the modeled area and exports it to a HEC-RAS terrain file.  DEM data
    may be sourced from any virtual raster source, but the default source is
    the `USGS 3DEP 1/3 arcsecond dataset
    <https://www.sciencebase.gov/catalog/item/4f70aa9fe4b058caae3f8de5>`_.  By
    default, terrain data is clipped to a 1,000-foot buffer around a concave
    hull of the submodel cross-sections, however, the buffer distance may be
    adjusted.  If resolution data is passed to the endpoint, the terrain raster
    will be resampled to that resolution.
    """
    logging.info(f"create_ras_terrain starting")

    if resolution and not resolution_units:
        raise ValueError(
            f"'resolution' arg has been provided but 'resolution_units' arg has not been provided. Please provide both"
        )

    if resolution_units:
        if resolution_units not in ["Feet", "Meters"]:
            raise ValueError(f"invalid resolution_units: {resolution_units}. expected 'Feet' or 'Meters'")

    nwm_rm = NwmReachModel(submodel_directory)

    if not nwm_rm.file_exists(nwm_rm.ras_gpkg_file):
        raise FileNotFoundError(f"NwmReachModel class expecting ras_gpkg_file {nwm_rm.ras_gpkg_file}, file not found")

    if not os.path.exists(nwm_rm.terrain_directory):
        os.makedirs(nwm_rm.terrain_directory, exist_ok=True)

    gdf_xs = gpd.read_file(nwm_rm.ras_gpkg_file, layer="XS").explode(ignore_index=True)
    crs = gdf_xs.crs

    mask = get_geometry_mask(nwm_rm.xs_concave_hull, terrain_source_url)

    # clip dem
    src_dem_clipped_localfile = os.path.join(nwm_rm.terrain_directory, "temp.tif")
    map_dem_clipped_basename = os.path.basename(terrain_source_url)
    src_dem_reprojected_localfile = os.path.join(
        nwm_rm.terrain_directory, map_dem_clipped_basename.replace(".vrt", ".tif")
    )

    clip_raster(
        terrain_source_url,
        src_dem_clipped_localfile,
        mask,
        vertical_units,
    )

    # reproject/resample dem
    logging.debug(f"Reprojecting/Resampling DEM {src_dem_clipped_localfile} to {src_dem_clipped_localfile}")
    reproject_raster(
        src_dem_clipped_localfile, src_dem_reprojected_localfile, CRS(nwm_rm.crs), resolution, resolution_units
    )
    os.remove(src_dem_clipped_localfile)

    # write projection file
    projection_file = write_projection_file(nwm_rm.crs, nwm_rm.terrain_directory)

    # Make the RAS mapping terrain locally
    result = create_terrain(
        [src_dem_reprojected_localfile],
        projection_file,
        f"{nwm_rm.terrain_directory}\\{nwm_rm.model_name}",
        vertical_units=MAP_DEM_VERT_UNITS,
    )
    os.remove(src_dem_reprojected_localfile)
    nwm_rm.update_write_ripple1d_parameters({"source_terrain": terrain_source_url})
    logging.info(f"create_ras_terrain complete")
    return result



"""Adds RAS Items to a STAC collection.

TODO: Refactor, this script was written in one setting and needs to be revised significantly.
"""

# import argparse
# import logging
# import os
# from typing import List

# import pystac
# from dotenv import load_dotenv

# from ripple1d.utils.s3_utils import init_s3_resources, list_keys
# from ripple1d.utils.stac_utils import (
#     create_collection,
#     key_to_uri,
#     upsert_collection,
#     upsert_item,
# )

# load_dotenv()
# # STAC API URL
# API_URL = os.getenv("STAC_API_URL")


# def main(
#     new_collection_items: List[pystac.Item],
#     collection_id: str,
#     description: str,
#     title: str,
# ):
#     """Given a list of items autogenerated using MCAT-RAS and ras-stac, this function creates a new collection in the STAC_API and adds the items to the collection."""
#     collection = create_collection(new_collection_items, collection_id, description, title)
#     r = upsert_collection(API_URL, collection)

#     for item in new_collection_items:
#         r = upsert_item(API_URL, collection_id, item)
#         logging.info(f"Adding {item.id} to {collection_id} collection")


# if __name__ == "__main__":

#     _, s3_client, _ = init_s3_resources()

#     parser = argparse.ArgumentParser(description="")
#     parser.add_argument("--collection_id", type=str, required=True, help="Collection ID")
#     parser.add_argument("--bucket_name", type=str, required=True, help="Bucket name")
#     parser.add_argument("--bucket_prefix", type=str, required=True, help="Bucket prefix")
#     parser.add_argument("--description", type=str, required=True, help="Description")
#     parser.add_argument("--title", type=str, required=True, help="Title")

#     args = parser.parse_args()

#     _, s3_client, _ = init_s3_resources()

#     collection_id = args.collection_id

#     bucket_name = args.bucket_name
#     bucket_prefix = args.bucket_prefix
#     title = args.title
#     description = args.description

#     # Read json files created using ras-stac and stored on s3
#     json_keys = list_keys(s3_client, bucket_name, bucket_prefix, "json")

#     new_collection_items = []

#     for jkey in json_keys[10:15]:
#         item = pystac.Item.from_file(key_to_uri(jkey, bucket_name))
#         new_collection_items.append(item)

#     main(new_collection_items, collection_id, description, title)



"""
Script is a prototype for creating a STAC collection for the FIM depth grids.

TODO: Refactor, this script was written in one setting and needs to be revised significantly.
"""

# from typing import List

# import pystac
# from fim_collection import FIMCollection, FIMCollectionRasItem
# from ras_stac.utils.dg_utils import *
# from ras_stac.utils.dg_utils import create_depth_grid_item
# from ras_stac.utils.ras_stac import *
# from ras_stac.utils.s3_utils import *
# from s3_utils import check_s3_key_exists, init_s3_resources, list_keys
# from stac_utils import create_collection, upsert_collection, upsert_item, uri_to_key


# def fim_dg_item(
#     dg_id: str,
#     dgs: list,
#     asset_list: list = None,
#     minio_mode: bool = False,
#     bucket_name: str = "fim",
# ):

#     # Instantitate S3 resources
#     session, _, s3_resource = init_s3_resources()
#     bucket = s3_resource.Bucket(bucket_name)
#     AWS_SESSION = AWSSession(session)

#     dg_obj = bucket.Object(dgs[0])
#     dg_item = create_depth_grid_item(dg_obj, dg_id, AWS_SESSION, minio_mode=minio_mode)

#     assets = []
#     for asset in dg_item.assets:
#         assets.append(asset)
#     for asset in assets:
#         del dg_item.assets[asset]

#     for asset_key in asset_list:
#         obj = bucket.Object(asset_key)
#         metadata = get_basic_object_metadata(obj)
#         asset = pystac.Asset(
#             f"https://fim.s3.amazonaws.com/{asset_key}",
#             extra_fields=metadata,
#             roles=["application/x-sqlite3"],
#             description="Additional data for the FIM DG item",
#         )
#         dg_item.add_asset(Path(asset_key).name.replace(" ", ""), asset)

#     for asset_key in dgs:
#         obj = bucket.Object(asset_key)
#         metadata = get_basic_object_metadata(obj)
#         asset = pystac.Asset(
#             f"https://fim.s3.amazonaws.com/{asset_key}",
#             extra_fields=metadata,
#             roles=["fim-dg", pystac.MediaType.COG],
#             description="(Prototype) RAS1D Fim Depth Grids",
#         )
#         dg_item.add_asset(f"{Path(asset_key).parent.name}-{Path(asset_key).name}", asset)

#     return dg_item


# def split_parts(key, prefix):
#     key_without_prefix = key[len(prefix) :].lstrip("/")
#     parts = key_without_prefix.split("/")
#     key_part = parts[0]
#     z_parts = parts[1:]
#     return {key_part: z_parts}


# def map_output_fims(bucket, output_prefix, client):
#     dg_stages = {}
#     tifs = list_keys(client, bucket, output_prefix, ".tif")

#     for t in tifs:
#         split_dict = split_parts(t, output_prefix)
#         for key_part, z_parts in split_dict.items():
#             if key_part not in dg_stages:
#                 dg_stages[key_part] = {}
#             z_key = z_parts[0]
#             if z_key not in dg_stages[key_part]:
#                 dg_stages[key_part][z_key] = []
#             dg_stages[key_part][z_key].append(z_parts[1])
#     return dg_stages


# def main(
#     new_collection_items: List[pystac.Item],
#     collection_id: str,
#     description: str,
#     title: str,
# ):
#     """
#     Given a list of items created using ras-stac library prior to publishing. This function
#     creates a new collection in the STAC_API and adds the items to the collection.
#     """
#     collection = create_collection(new_collection_items, collection_id, description, title)
#     r = upsert_collection(API_URL, collection)
#     logging.info(f"response: {r}")

#     for item in new_collection_items:
#         r = upsert_item(API_URL, collection_id, item)
#         logging.info(f"item_id: {item.id},response: {r}")


# if __name__ == "__main__":

#     _, s3_client, _ = init_s3_resources()

#     huc_id = "12040101"
#     ras_collection_id = f"huc-{huc_id}"
#     fim_collection_id = f"huc-{huc_id}-fims"

#     bucket_name = "fim"
#     bucket_prefix = "stac/12040101"

#     description = f"""Prototype catalog for FIMS from the huc-{huc_id} RAS model collection"""
#     title = f"FIMS for HUC-{huc_id}"

#     # STAC API URL
#     API_URL = "https://stac2.dewberryanalytics.com"

#     fc = FIMCollection(API_URL, ras_collection_id)
#     new_dg_items = []
#     i = 0
#     for item in fc.collection.get_all_items():
#         i += 1
#         logging.info(f"{i} : {item.id}")
#         try:
#             model_branches = item.properties["FIM:Branch Metadata"].keys()
#             # logging.debug(f"{len(model_branches)} branches in FIM:Branch Metadata for {item.id}")
#         except KeyError:
#             logging.error(f"FIM:Branch Metadata does not exist for {item.id}")

#         prefix = f"mip/dev/ripple1d/output/collections/{ras_collection_id}/items/{item.id}"
#         model_db = f"{prefix}/{item.id}.db"
#         ripple1d_succeed = f"{prefix}/ripple1d-succeed.json"

#         if check_s3_key_exists("fim", ripple1d_succeed):
#             # logging.debug(f"Model database exists for {item.id}")
#             fim_dict = map_output_fims(bucket_name, f"{prefix}/", s3_client)
#             fci = FIMCollectionRasItem(API_URL, ras_collection_id, item.id)

#             for branch_id, stage_values in fim_dict.items():
#                 dg_id = branch_id
#                 assets = [model_db]
#                 dgs = []
#                 for z_key in stage_values.keys():
#                     dgs.extend([f"{prefix}/{branch_id}/{z_key}/{v}" for v in fim_dict[branch_id][z_key]])
#                 dg_item = fim_dg_item(dg_id, dgs, assets)
#                 dg_item.add_derived_from(item)

#                 for asset_name in dg_item.assets:
#                     asset = dg_item.assets[asset_name]
#                     asset.extra_fields["s3_key"] = uri_to_key(asset.href, bucket_name)
#                 dg_item.properties["FIM:Branch Metadata"] = item.properties["FIM:Branch Metadata"][branch_id]

#                 dg_item.properties["software"] = "ripple1d v0.1.0-alpha.1"

#                 new_dg_items.append(dg_item)

#                 # dg_item_link = pystac.Link(
#                 #         rel="related",
#                 #         target=dg_item,
#                 #         title=f"FIMs-{branch_id}",
#                 #     )

#     main(new_dg_items, fim_collection_id, description, title)



"""Logging utility an setup."""

import inspect
import json
import logging
import os
import time
import traceback
from datetime import datetime, timezone
from logging.handlers import RotatingFileHandler

from ripple1d.consts import SUPPRESS_LOGS


class RippleLogFormatter(logging.Formatter):
    """Format log messages as JSON-LD."""

    def __init__(
        self, datefmt="%Y-%m-%dT%H:%M:%SZ", include_traceback: bool = True, include_fields: list = None, **kwargs
    ):
        """Initialize the formatter with options."""
        super().__init__(datefmt=datefmt, **kwargs)
        self.include_traceback = include_traceback

        # List of valid field options
        self.include_fields_options = [
            "logger_name",
            "function_name",
            "line_number",
            "filename",
            "thread_name",
            "process_name",
            "process_id",
            "error",
            "traceback",
        ]

        # Validate and set include_fields
        if include_fields is None:
            self.include_fields = []
        else:
            self.include_fields = [field for field in include_fields if field in self.include_fields_options]

    def format(self, record):
        """Set up fine-grain format and logging options."""
        # Get the function name of the caller
        stack = inspect.stack()
        for frame_info in stack[1:]:
            if frame_info.function != "log":
                record.funcName = frame_info.function
                break

        # Basic log entry fields that are always included
        log_entry = {
            "@type": "RippleLogs",
            "timestamp": self.formatTime(record, self.datefmt),
            "level": record.levelname,
            "msg": record.getMessage(),
        }

        # Conditionally add fields based on the include_fields list
        if "logger_name" in self.include_fields:
            log_entry["logger_name"] = record.name
        if "function_name" in self.include_fields:
            log_entry["function_name"] = record.funcName
        if "line_number" in self.include_fields:
            log_entry["line_number"] = record.lineno
        if "filename" in self.include_fields:
            log_entry["filename"] = record.pathname
        if "thread_name" in self.include_fields:
            log_entry["thread_name"] = record.threadName
        if "process_name" in self.include_fields:
            log_entry["process_name"] = record.processName
        if "process_id" in self.include_fields:
            log_entry["process_id"] = record.process
        if "error" in self.include_fields:
            log_entry["error"] = None
        if "traceback" in self.include_fields and record.exc_info and self.include_traceback:
            log_entry["error"] = str(record.exc_info[1])
            log_entry["traceback"] = traceback.format_exc()

        return json.dumps(log_entry)


def configure_logging(level, logfile: str = None, milliseconds: bool = False, verbose: bool = False):
    """Configure logging for ripple1d."""
    if milliseconds:
        datefmt = "%Y-%m-%dT%H:%M:%S.%fZ"
    else:
        datefmt = "%Y-%m-%dT%H:%M:%SZ"

    if verbose:
        include_fields = [
            "logger_name",
            "function_name",
            "line_number",
            "filename",
            "thread_name",
            "process_name",
            "process_id",
        ]
    else:
        include_fields = []

    log_formatter = RippleLogFormatter(datefmt=datefmt, include_traceback=True, include_fields=include_fields)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    handlers = [console_handler]

    if logfile:
        file_handler = RotatingFileHandler(logfile, maxBytes=5000000, backupCount=5)
        file_handler.setFormatter(log_formatter)
        handlers.append(file_handler)

    for module in SUPPRESS_LOGS:
        logging.getLogger(module).setLevel(logging.WARNING)

    logging.basicConfig(
        level=level,
        handlers=handlers,
    )


def initialize_server_logger(log_dir: str = "", log_level: int = logging.INFO) -> logging.Logger:
    """Initialize log with JSON-LD style formatting and throttled level for AWS libs.

    By default sends to StreamHandler (stdout/stderr), but can provide a filename to log to disk instead.
    """
    filename = os.path.join(log_dir, "server-logs.jsonld")

    for module in SUPPRESS_LOGS:
        logging.getLogger(module).setLevel(logging.ERROR)

    log = logging.getLogger()
    log.setLevel(log_level)
    formatter = RippleLogFormatter()

    if filename:
        print(f"Initializing log file: {filename}")
        os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
        file_handler = logging.FileHandler(filename=filename)
        file_handler.setFormatter(formatter)
        log.addHandler(file_handler)
    else:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        log.addHandler(stream_handler)

    return log



"""Subset a geopackage based on clonfation with NWM hydrofabric."""

import json
import logging
import os
import time
import warnings

import fiona
import geopandas as gpd
import pandas as pd
from shapely import LineString
from shapely.ops import split

import ripple1d
from ripple1d.consts import METERS_PER_FOOT
from ripple1d.data_model import NwmReachModel, RippleSourceDirectory, RippleSourceModel
from ripple1d.utils.ripple_utils import (
    clip_ras_centerline,
    fix_reversed_xs,
    xs_concave_hull,
)

warnings.filterwarnings("ignore")


class RippleGeopackageSubsetter:
    """Subset a geopackage based on conflation with NWM hydrofabric."""

    def __init__(self, src_gpkg_path: str, conflation_json: str, dst_project_dir: str, nwm_id: str = None):

        self.src_gpkg_path = src_gpkg_path
        self.conflation_json = conflation_json
        self.dst_project_dir = dst_project_dir
        self.nwm_id = nwm_id
        self._subset_gdf = None
        self._source_junction = None
        self._source_river = None
        self._source_xs = None
        self._source_structure = None
        self._conflation_parameters = None
        self._source_hulls = None
        self._ripple_xs_concave_hull = None

    def set_nwm_id(self, nwm_id: str):
        """Set the network ID."""
        self.nwm_id = nwm_id
        self._subset_gdf = None
        self._ripple_xs_concave_hull = None

    @property
    def ripple_us_xs(self):
        """The most upstream cross section for the ripple model."""
        return self.ripple_xs.loc[
            self.ripple_xs["river_station"] == self.ripple_xs["river_station"].max(), "geometry"
        ].iloc[0]

    @property
    def ripple_ds_xs(self):
        """The most downstream cross section for the ripple model."""
        return self.ripple_xs.loc[
            self.ripple_xs["river_station"] == self.ripple_xs["river_station"].min(), "geometry"
        ].iloc[0]

    @property
    def split_source_hull(self):
        """Split the concave hull of the source model using the upstream and downstream cross sections of the submodel."""
        geoms = split(self.source_hulls.geometry.iloc[0], self.ripple_us_xs).geoms
        hulls = []
        for geom in geoms:
            if geom.intersects(self.ripple_us_xs) and geom.intersects(self.ripple_ds_xs):
                candidate_geoms = split(geom, self.ripple_ds_xs).geoms
                for candidate_geom in candidate_geoms:
                    if candidate_geom.intersects(self.ripple_us_xs) and candidate_geom.intersects(self.ripple_ds_xs):
                        hulls.append(candidate_geom)
        if len(hulls) != 1:
            raise ValueError(
                f"Expected 1 polygon for ripple xs concave hull; got: {len(hulls)} | network id: {self.nwm_id}"
            )
        return hulls

    @property
    def ripple_xs_concave_hull(self):
        """Get the concave hull of the cross sections."""
        if self._ripple_xs_concave_hull is None:
            try:
                hulls = self.split_source_hull
                self._ripple_xs_concave_hull = gpd.GeoDataFrame({"geometry": hulls}, geometry="geometry", crs=self.crs)
            except Exception as e:
                self._ripple_xs_concave_hull = xs_concave_hull(fix_reversed_xs(self.ripple_xs, self.ripple_river))

        return self._ripple_xs_concave_hull

    @property
    def conflation_parameters(self) -> dict:
        """Extract conflation parameters from the conflation json."""
        if self._conflation_parameters is None:
            with open(self.conflation_json, "r") as f:
                self._conflation_parameters = json.load(f)
        return self._conflation_parameters

    @property
    def ripple1d_parameters(self) -> dict:
        """Extract ripple1d parameters from the conflation json."""
        return self.conflation_parameters["reaches"][self.nwm_id]

    @property
    def us_reach(self) -> str:
        """Extract upstream reach from conflation parameters."""
        return self.ripple1d_parameters["us_xs"]["reach"]

    @property
    def us_river(self) -> str:
        """Extract upstream river from conflation parameters."""
        return self.ripple1d_parameters["us_xs"]["river"]

    @property
    def us_rs(self) -> str:
        """Extract upstream river station from conflation parameters."""
        return self.ripple1d_parameters["us_xs"]["xs_id"]

    @property
    def ds_river(self) -> str:
        """Extract downstream river from conflation parameters."""
        return self.ripple1d_parameters["ds_xs"]["river"]

    @property
    def ds_reach(self) -> str:
        """Extract downstream reach from conflation parameters."""
        return self.ripple1d_parameters["ds_xs"]["reach"]

    @property
    def ds_rs(self) -> str:
        """Extract downstream river station from conflation parameters."""
        return self.ripple1d_parameters["ds_xs"]["xs_id"]

    @property
    def source_xs(self) -> gpd.GeoDataFrame:
        """Extract cross sections from the source geopackage."""
        if self._source_xs is None:
            xs = gpd.read_file(self.src_gpkg_path, layer="XS")
            self._source_xs = xs[xs.intersects(self.source_river.union_all())]
        return self._source_xs

    @property
    def source_hulls(self) -> gpd.GeoDataFrame:
        """Extract cross sections from the source geopackage."""
        if self._source_hulls is None:
            # TODO we can read the concave hull of the source model from the gpkg once we decide that it is required layer.
            # self._source_hulls = gpd.read_file(self.src_gpkg_path, layer="XS_concave_hull")
            self._source_hulls = xs_concave_hull(
                fix_reversed_xs(self.source_xs, self.source_river), self.source_junction
            )
        return self._source_hulls

    @property
    def source_river(self) -> gpd.GeoDataFrame:
        """Extract river geometry from the source geopackage."""
        if self._source_river is None:
            self._source_river = gpd.read_file(self.src_gpkg_path, layer="River")
        return self._source_river

    @property
    def source_structure(self) -> gpd.GeoDataFrame:
        """Extract structures from the source geopackage."""
        if "Structure" in fiona.listlayers(self.src_gpkg_path):
            if self._source_structure is None:
                structures = gpd.read_file(self.src_gpkg_path, layer="Structure")
                self._source_structure = structures[structures.intersects(self.source_river.union_all())]
            return self._source_structure

    @property
    def source_junction(self) -> gpd.GeoDataFrame:
        """Extract junctions from the source geopackage."""
        if "Junction" in fiona.listlayers(self.src_gpkg_path):
            if self._source_junction is None:
                self._source_junction = gpd.read_file(self.src_gpkg_path, layer="Junction")
            return self._source_junction

    @property
    def ripple_xs(self) -> gpd.GeoDataFrame:
        """Subset cross sections based on NWM reach."""
        return self.subset_gdfs["XS"]

    @property
    def ripple_river(self) -> gpd.GeoDataFrame:
        """Subset river geometry based on NWM reach."""
        return self.subset_gdfs["River"]

    @property
    def ripple_structure(self) -> gpd.GeoDataFrame:
        """Subset structures based on NWM reach."""
        return self.subset_gdfs["Structure"]

    @property
    def subset_gdfs(self) -> dict:
        """Subset the cross sections, structues, and river geometry for a given NWM reach."""
        if self._subset_gdf is None:
            # subset data
            if self.us_river == self.ds_river and self.us_reach == self.ds_reach:
                ripple_xs, ripple_structure, ripple_river = self.process_as_one_ras_reach()
            else:
                ripple_xs, ripple_structure, ripple_river = self.process_as_multiple_ras_reach()

            # check if only 1 cross section for nwm_reach
            if len(ripple_xs) <= 1:
                logging.warning(f"Only 1 cross section conflated to NWM reach {self.nwm_id}. Skipping this reach.")
                return None

            # update fields
            ripple_xs = self.update_fields(ripple_xs)
            ripple_river = self.update_fields(ripple_river)
            ripple_structure = self.update_fields(ripple_structure)

            # clip river to cross sections
            ripple_river["geometry"] = clip_ras_centerline(
                ripple_river.iloc[0].geometry, fix_reversed_xs(ripple_xs, ripple_river), 2
            )

            if ripple_structure is not None and len(ripple_structure) > 0:
                self._subset_gdf = {
                    "XS": ripple_xs,
                    "River": ripple_river,
                    "Structure": ripple_structure,
                }
            else:
                self._subset_gdf = {
                    "XS": ripple_xs,
                    "River": ripple_river,
                }

        return self._subset_gdf

    @property
    def ripple_gpkg_file(self) -> str:
        """Return the path to the new geopackage."""
        return self.nwm_reach_model.ras_gpkg_file

    @property
    def nwm_reach_model(self) -> NwmReachModel:
        """Return the new NWM reach model object."""
        return NwmReachModel(self.dst_project_dir)

    def write_ripple_gpkg(self) -> None:
        """Write the subsetted geopackage to the destination project directory."""
        os.makedirs(self.dst_project_dir, exist_ok=True)

        for layer, gdf in self.subset_gdfs.items():
            # remove lateral structures
            if layer == "Structure":
                if (gdf["type"] == 6).any():
                    logging.warning(
                        f"Lateral structures are not currently supported in ripple1d. The lateral structures will be dropped."
                    )
                    gdf = gdf.loc[gdf["type"] != 6, :]

            if gdf.shape[0] > 0:
                gdf.to_file(self.ripple_gpkg_file, layer=layer)
                if layer == "XS":
                    self.ripple_xs_concave_hull.to_file(self.ripple_gpkg_file, driver="GPKG", layer="XS_concave_hull")

    @property
    def min_flow(self) -> float:
        """Extract the min flow from the cross sections."""
        if "flows" in self.ripple_xs.columns:
            return self.ripple_xs["flows"].str.split("\n", expand=True).astype(float).min().min()
        else:
            logging.warning(f"no flows specified in source model gpkg for {self.nwm_id}")
            return 10000000000000

    @property
    def max_flow(self) -> float:
        """Extract the max flow from the cross sections."""
        if "flows" in self.ripple_xs.columns:
            return self.ripple_xs["flows"].str.split("\n", expand=True).astype(float).max().max()
        else:
            logging.warning(f"no flows specified in source model gpkg for {self.nwm_id}")
            return 0

    @property
    def crs(self):
        """Extract the CRS from the cross sections."""
        return self.source_xs.crs

    def walk_junctions(self) -> list[str]:
        """Check if junctions are present for the given river-reaches."""
        river_reaches = []
        if not self.source_junction.empty:
            c = 0
            while True:
                c += 1
                if c > 100:
                    logging.warning(
                        f"Could not find junction for: {self.nwm_id}. The reach may contain cross sections from multiple RAS reaches that are not connected via junctions."
                    )
                    return None

                for _, row in self.source_junction.iterrows():

                    us_rivers = row.us_rivers.split(",")
                    us_reaches = row.us_reaches.split(",")

                    for river, reach in zip(us_rivers, us_reaches):
                        if row.ds_rivers == self.ds_river and row.ds_reaches == self.ds_reach:
                            return river_reaches
                        if river == self.us_river and reach == self.us_reach:
                            river_reaches.append(f"{row.ds_rivers.ljust(16)},{row.ds_reaches.ljust(16)}")
                            us_river = row.ds_rivers
                            us_reach = row.ds_reaches

    def clean_river_stations(self, ras_data: str) -> str:
        """Clean up river station data."""
        lines = ras_data.splitlines()
        data = lines[0].split(",")
        if "*" in data[1]:
            data[1] = str(float(data[1].rstrip("*"))) + "*"
            data[1] = data[1].ljust(8)
        else:
            data[1] = str(float(data[1])).ljust(8)
        lines[0] = ",".join(data)
        return "\n".join(lines) + "\n"

    def round_river_stations(self, ras_data: str) -> str:
        """Clean up river station data."""
        lines = ras_data.splitlines()
        data = lines[0].split(",")
        if "*" in data[1]:
            data[1] = str(float(round(float(data[1].rstrip("*"))))) + "*"
            data[1] = data[1].ljust(8)
        else:
            data[1] = str(float(round(float(data[1])))).ljust(8)
        lines[0] = ",".join(data)
        return "\n".join(lines) + "\n"

    def update_river_station(self, ras_data: str, river_station: str) -> str:
        """Update river station data."""
        lines = ras_data.splitlines()
        data = lines[0].split(",")
        if "*" in data[1]:
            data[1] = str(float(data[1].rstrip("*")) + river_station) + "*"
            data[1] = data[1].ljust(8)
        else:
            data[1] = str(float(data[1]) + river_station).ljust(8)
        lines[0] = ",".join(data)
        return "\n".join(lines) + "\n"

    def junction_length_to_reach_lengths(self):
        """Adjust reach lengths using junction."""
        # TODO adjust reach lengths using junction lengths
        raise NotImplementedError
        # for row in junction_gdf.iterrows():
        #     if us_river in row["us_rivers"] and us_reach in row["us_reach"]:

    def process_as_one_ras_reach(self) -> tuple:
        """Process as a single ras-river-reach."""
        xs_subset_gdf = self.source_xs.loc[
            (self.source_xs["river"] == self.us_river)
            & (self.source_xs["reach"] == self.us_reach)
            & (self.source_xs["river_station"] >= float(self.ds_rs))
            & (self.source_xs["river_station"] <= float(self.us_rs))
        ]
        if self.source_structure is not None:
            structures_subset_gdf = self.source_structure.loc[
                (self.source_structure["river"] == self.us_river)
                & (self.source_structure["reach"] == self.us_reach)
                & (self.source_structure["river_station"] >= float(self.ds_rs))
                & (self.source_structure["river_station"] <= float(self.us_rs))
            ]
        else:
            structures_subset_gdf = None
        river_subset_gdf = self.source_river.loc[
            (self.source_river["river"] == self.us_river) & (self.source_river["reach"] == self.us_reach)
        ]

        return xs_subset_gdf, structures_subset_gdf, river_subset_gdf

    @property
    def xs_us_reach(self) -> gpd.GeoDataFrame:
        """Extract cross sections for the upstream reach."""
        return self.source_xs.loc[
            (self.source_xs["river"] == self.us_river)
            & (self.source_xs["reach"] == self.us_reach)
            & (self.source_xs["river_station"] <= float(self.us_rs))
        ]

    @property
    def xs_ds_reach(self) -> gpd.GeoDataFrame:
        """Extract cross sections for the downstream reach."""
        return self.source_xs.loc[
            (self.source_xs["river"] == self.ds_river)
            & (self.source_xs["reach"] == self.ds_reach)
            & (self.source_xs["river_station"] >= float(self.ds_rs))
        ]

    @property
    def structures_us_reach(self) -> gpd.GeoDataFrame:
        """Extract structures for the upstream reach."""
        if self.source_structure is not None:
            return self.source_structure.loc[
                (self.source_structure["river"] == self.us_river)
                & (self.source_structure["reach"] == self.us_reach)
                & (self.source_structure["river_station"] <= float(self.us_rs))
            ]

    @property
    def structures_ds_reach(self) -> gpd.GeoDataFrame:
        """Extract structures for the downstream reach."""
        if self.source_structure is not None:
            return self.source_structure.loc[
                (self.source_structure["river"] == self.ds_river)
                & (self.source_structure["reach"] == self.ds_reach)
                & (self.source_structure["river_station"] >= float(self.ds_rs))
            ]

    def add_intermediate_river_reaches(self, intermediate_river_reaches) -> gpd.GeoDataFrame:
        """Add intermediate river reaches to the xs_us_reach."""
        xs_us_reach = self.xs_us_reach.copy()
        for intermediate_river_reach in intermediate_river_reaches:
            xs_intermediate_river_reach = self.source_xs.loc[self.source_xs["river_reach"] == intermediate_river_reach]
            if xs_us_reach["river_station"].min() <= xs_intermediate_river_reach["river_station"].max():
                logging.info(
                    f"The lowest river station on the upstream reach ({xs_us_reach['river_station'].min()}) is less"
                    f" than the highest river station on the intermediate reach"
                    f"({xs_intermediate_river_reach['river_station'].max()}) for nwm_id: {self.nwm_id}. The river"
                    f" stationing of the upstream reach will be updated to ensure river stationings increase from"
                    "downstream to upstream"
                )
                xs_us_reach["river_station"] = (
                    xs_us_reach["river_station"] + xs_intermediate_river_reach["river_station"].max()
                )
                xs_us_reach["ras_data"] = xs_us_reach["ras_data"].apply(
                    lambda ras_data: self.update_river_station(
                        ras_data, xs_intermediate_river_reach["river_station"].max()
                    )
                )
            xs_us_reach = pd.concat([xs_us_reach, xs_intermediate_river_reach])
        return xs_us_reach

    def adjust_river_stations(self, xs_us_reach, structures_us_reach) -> tuple:
        """Adjust river stations of the upstream reach if the min river station of the upstream reach is less than the max river station of the downstream reach."""
        if xs_us_reach["river_station"].min() <= self.xs_ds_reach["river_station"].max():
            logging.info(
                f"the lowest river station on the upstream reach ({xs_us_reach['river_station'].min()}) is less"
                f" than the highest river station on the downstream reach ({self.xs_ds_reach['river_station'].max()}) for nwm_id: {self.nwm_id}"
            )
            xs_us_reach["river_station"] = xs_us_reach["river_station"] + self.xs_ds_reach["river_station"].max()
            xs_us_reach["ras_data"] = xs_us_reach["ras_data"].apply(
                lambda ras_data: self.update_river_station(ras_data, self.xs_ds_reach["river_station"].max())
            )
            if structures_us_reach is not None:
                structures_us_reach["river_station"] = (
                    structures_us_reach["river_station"] + self.xs_ds_reach["river_station"].max()
                )
                structures_us_reach["ras_data"] = structures_us_reach["ras_data"].apply(
                    lambda ras_data: self.update_river_station(ras_data, self.xs_ds_reach["river_station"].max())
                )
        return xs_us_reach, structures_us_reach

    def process_as_multiple_ras_reach(self) -> tuple:
        """Process as multiple ras-river-reach."""
        if "Junction" in fiona.listlayers(self.src_gpkg_path):
            junction_gdf = gpd.read_file(self.src_gpkg_path, layer="Junction")
            intermediate_river_reaches = self.walk_junctions()
        else:
            intermediate_river_reaches = None

        # add intermediate river reaches to the upstream reach
        if intermediate_river_reaches:
            xs_us_reach = self.add_intermediate_river_reaches(intermediate_river_reaches)
        else:
            xs_us_reach = self.xs_us_reach.copy()

        # update river stations
        xs_us_reach, structures_us_reach = self.adjust_river_stations(xs_us_reach, self.structures_us_reach)
        # combine us and ds gdfs
        xs_subset_gdf = pd.concat([xs_us_reach, self.xs_ds_reach])
        river_subset_gdf = self.combine_reach_features(intermediate_river_reaches)

        if self.source_structure is not None:
            structures_subset_gdf = pd.concat([structures_us_reach, self.structures_ds_reach])
        else:
            structures_subset_gdf = None

        return xs_subset_gdf, structures_subset_gdf, river_subset_gdf

    def rename_river_reach(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Rename river, reach, and river_reach columns after the nwm reach id."""
        pd.options.mode.copy_on_write = True
        gdf["river"] = self.nwm_id
        gdf["reach"] = self.nwm_id
        gdf["river_reach"] = f"{self.nwm_id.ljust(16)},{self.nwm_id.ljust(16)}"

        return gdf

    def combine_reach_features(self, intermediate_river_reaches: list[str]) -> gpd.GeoDataFrame:
        """Combine reach coordinates and update river and reach names to be nwm id."""
        us_reach = self.source_river.loc[
            (self.source_river["river"] == self.us_river) & (self.source_river["reach"] == self.us_reach)
        ]
        ds_reach = self.source_river.loc[
            (self.source_river["river"] == self.ds_river) & (self.source_river["reach"] == self.ds_reach)
        ]

        # handle river reach coords
        coords = list(us_reach.iloc[0]["geometry"].coords)
        if intermediate_river_reaches:
            for intermediate_river_reach in intermediate_river_reaches:
                intermediate_river_reach_reach = self.source_river.loc[
                    self.source_river["river_reach"] == intermediate_river_reach
                ]
                coords += list(intermediate_river_reach_reach.iloc[0]["geometry"].coords)
        coords += list(ds_reach.iloc[0]["geometry"].coords)

        return gpd.GeoDataFrame(
            {"geometry": [LineString(coords)], "river": [self.nwm_id], "reach": [self.nwm_id]},
            geometry="geometry",
            crs=self.crs,
        )

    def update_fields(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Update fields for the new NWM reach."""
        if gdf is not None:
            gdf = self.rename_river_reach(gdf)

            # clean river stations
            if "river_station" in gdf.columns:
                if (gdf["river_station"].astype(str).str.len() > 8).any():
                    gdf["ras_data"] = gdf["ras_data"].apply(lambda ras_data: self.round_river_stations(ras_data))
                    gdf["river_station"] = gdf["river_station"].round().astype(float)
                else:
                    gdf["ras_data"] = gdf["ras_data"].apply(lambda ras_data: self.clean_river_stations(ras_data))
                    gdf["river_station"] = gdf["river_station"].astype(float)
        return gdf

    def update_ripple1d_parameters(self, rsd: RippleSourceDirectory):
        """Update ripple1d_parameters with results of subsetting."""
        ripple1d_parameters = self.ripple1d_parameters
        ripple1d_parameters["source_model"] = rsd.ras_project_file
        ripple1d_parameters["crs"] = self.crs.to_wkt()
        ripple1d_parameters["version"] = ripple1d.__version__
        ripple1d_parameters["high_flow"] = max([ripple1d_parameters["high_flow"], self.max_flow])
        ripple1d_parameters["low_flow"] = min([ripple1d_parameters["low_flow"], self.min_flow])
        if ripple1d_parameters["high_flow"] == self.max_flow:
            ripple1d_parameters["notes"] = ["high_flow computed from source model flows"]
        if ripple1d_parameters["low_flow"] == self.min_flow:
            ripple1d_parameters["notes"] = ["low_flow computed from source model flows"]
        return ripple1d_parameters

    def write_ripple1d_parameters(self, ripple1d_parameters: dict):
        """Write ripple1d parameters to json file."""
        with open(os.path.join(self.dst_project_dir, f"{self.nwm_id}.ripple1d.json"), "w") as f:
            json.dump(ripple1d_parameters, f, indent=4)


def extract_submodel(source_model_directory: str, submodel_directory: str, nwm_id: int):
    """Use ripple conflation data to create a new GPKG from an existing ras geopackage.

    Create a new geopackage with information for a specific NWM reach.  The new geopackage contains layer for the river centerline, cross-sections, and structures.

    Parameters
    ----------
    source_model_directory : str
        The path to the directory containing HEC-RAS project, plan, geometry, and flow files.
    submodel_directory : str
        The path to export submodel HEC-RAS files to.
    nwm_id : int
        The id of the NWM reach to create a submodel for
    task_id : str, optional
        Task ID to use for logging, by default ""

    Returns
    -------
    dict
        Metadata for the submodel

    Raises
    ------
    FileNotFoundError
        Raised when no geopackage is found in the source model directory
    FileNotFoundError
        Raised when no .conflation.json is found in the source model directory
    """
    # time.sleep(10)

    if not os.path.exists(source_model_directory):
        raise FileNotFoundError(
            f"cannot find directory for source model {source_model_directory}, please ensure dir exists"
        )
    rsd = RippleSourceDirectory(source_model_directory)

    logging.info(f"extract_submodel starting for nwm_id {nwm_id}")
    # print(f"preparing to extract NWM ID {nwm_id} from {os.path.basename(rsd.ras_project_file)}")

    if not rsd.file_exists(rsd.ras_gpkg_file):
        raise FileNotFoundError(f"cannot find file ras-geometry file {rsd.ras_gpkg_file}, please ensure file exists")

    if not rsd.file_exists(rsd.conflation_file):
        raise FileNotFoundError(f"cannot find conflation file {rsd.conflation_file}, please ensure file exists")

    ripple1d_parameters = rsd.nwm_conflation_parameters(str(nwm_id))
    if ripple1d_parameters["eclipsed"]:
        ripple1d_parameters["messages"] = f"skipping {nwm_id}; no cross sections conflated."
        logging.warning(ripple1d_parameters["messages"])
        gpkg_path = None

    else:
        rgs = RippleGeopackageSubsetter(rsd.ras_gpkg_file, rsd.conflation_file, submodel_directory, nwm_id)
        rgs.write_ripple_gpkg()
        ripple1d_parameters = rgs.update_ripple1d_parameters(rsd)
        rgs.write_ripple1d_parameters(ripple1d_parameters)
        gpkg_path = rgs.ripple_gpkg_file

    logging.info(f"extract_submodel complete for nwm_id {nwm_id}")
    return {"ripple1d_parameters": rsd.conflation_file, "ripple_gpkg_file": gpkg_path}



"""Utils for working with sqlite databases."""

import os
import sqlite3

import pandas as pd

from ripple1d.ras import RasManager


def create_db_and_table(db_name: str, table_name: str):
    """Create sqlite database and table."""
    os.makedirs(os.path.dirname(os.path.abspath(db_name)), exist_ok=True)
    sql_query = f"""
        CREATE TABLE {table_name}(
            reach_id INTEGER,
            ds_depth REAL,
            ds_wse REAL,
            us_flow INTEGER,
            us_depth REAL,
            us_wse REAL,
            boundary_condition TEXT, -- [kwse, nd]
            plan_suffix TEXT,
            map_exist BOOL CHECK(map_exist IN (0, 1)),
            UNIQUE(reach_id, us_flow, ds_wse, boundary_condition, plan_suffix)
        )
    """
    conn = sqlite3.connect(db_name)
    c = conn.cursor()
    c.execute(sql_query)
    conn.commit()
    conn.close()


def insert_data(
    db_name: str, table_name: str, data: pd.DataFrame, plan_suffix: str, missing_grids: list, boundary_condition: str
):
    """Insert data into the sqlite database."""
    conn = sqlite3.connect(db_name)
    c = conn.cursor()

    for row in data.itertuples():
        if boundary_condition == "kwse":
            if f"f_{int(row.us_flow)}-z_{str(row.ds_wse).replace(".","_")}" in missing_grids:
                map_exist = 0
            else:
                map_exist = 1
        elif boundary_condition == "nd":
            if str(int(row.us_flow)) in missing_grids:
                map_exist = 0
            else:
                map_exist = 1
        else:
            raise ValueError(
                f"Could not detemine boundary condition type for {boundary_condition}; expected kwse or nd"
            )

        c.execute(
            f"""
            INSERT OR REPLACE INTO {table_name} (reach_id, ds_depth, ds_wse, us_flow, us_depth, us_wse, boundary_condition,  plan_suffix, map_exist)
            VALUES ( ?, ?, ?, ?, ?, ?, ?, ?, ?)
        """,
            (
                int(row.reach_id),
                round(float(row.ds_depth), 1),
                round(float(row.ds_wse), 1),
                round(float(row.us_flow), 1),
                round(float(row.us_depth), 1),
                round(float(row.us_wse), 1),
                str(boundary_condition),
                str(plan_suffix),
                str(map_exist),
            ),
        )

    conn.commit()
    conn.close()


def parse_stage_flow(wses: pd.DataFrame) -> pd.DataFrame:
    """Parse flow and control by stage from profile names."""
    wses_t = wses.T
    wses_t.reset_index(inplace=True)
    wses_t[["us_flow", "ds_wse"]] = wses_t["index"].str.split("-", n=1, expand=True)
    wses_t.drop(columns="index", inplace=True)
    wses_t["us_flow"] = wses_t["us_flow"].str.lstrip("f_").astype(float)
    wses_t["ds_wse"] = wses_t["ds_wse"].str.lstrip("z_")
    wses_t["ds_wse"] = wses_t["ds_wse"].str.replace("_", ".").astype(float)

    return wses_t


def zero_depth_to_sqlite(
    rm: RasManager,
    plan_name: str,
    plan_suffix: str,
    nwm_id: str,
    missing_grids: list,
    database_path: str,
    table_name: str,
):
    """Export zero depth (normal depth) results to sqlite."""
    # set the plan
    rm.plan = rm.plans[plan_name]

    # read in flow/wse
    wses, flows = rm.plan.read_rating_curves()

    # get river-reach-rs
    us_river_reach_rs = rm.plan.geom.rivers[nwm_id][nwm_id].us_xs.river_reach_rs
    ds_river_reach_rs = rm.plan.geom.rivers[nwm_id][nwm_id].ds_xs.river_reach_rs

    wses_t = wses.T
    wses_t["us_flow"] = wses_t.index
    wses_t["ds_depth"] = 0
    df = wses_t.loc[:, [us_river_reach_rs, "us_flow", "ds_depth"]]
    df.rename(columns={us_river_reach_rs: "us_wse"}, inplace=True)

    # convert elvation to stage for upstream cross section
    us_thalweg = rm.plan.geom.rivers[nwm_id][nwm_id].us_xs.thalweg
    df["us_depth"] = df["us_wse"] - us_thalweg

    # convert elvation to stage for downstream cross section
    ds_df = wses_t.loc[:, [ds_river_reach_rs, "us_flow", "ds_depth"]]
    ds_df.rename(columns={ds_river_reach_rs: "us_wse"}, inplace=True)
    ds_thalweg = rm.plan.geom.rivers[nwm_id][nwm_id].ds_xs.thalweg

    df["ds_wse"] = ds_df["us_wse"]
    df["ds_depth"] = df["ds_wse"] - ds_thalweg

    # add control id
    df["reach_id"] = [nwm_id] * len(df)

    insert_data(database_path, table_name, df, plan_suffix, missing_grids, boundary_condition="nd")


def rating_curves_to_sqlite(
    rm: RasManager,
    plan_name: str,
    plan_suffix: str,
    nwm_id: str,
    missing_grids: list,
    database_path: str,
    table_name: str,
):
    """Export rating curves to sqlite."""
    # set the plan
    if plan_name not in rm.plans:
        return
    rm.plan = rm.plans[plan_name]

    # read in flow/wse
    wses, flows = rm.plan.read_rating_curves()

    # parse applied stage and flow from profile names
    wses = parse_stage_flow(wses)

    # get river-reach-rs id
    river_reach_rs = rm.plan.geom.rivers[nwm_id][nwm_id].us_xs.river_reach_rs

    # get subset of results for this cross section
    df = wses[["us_flow", "ds_wse", river_reach_rs]].copy()

    # rename columns
    df.rename(columns={river_reach_rs: "us_wse"}, inplace=True)

    # add control id
    df["reach_id"] = [nwm_id] * len(df)

    # convert elevation to stage
    thalweg = rm.plan.geom.rivers[nwm_id][nwm_id].us_xs.thalweg
    df["us_depth"] = df["us_wse"] - thalweg

    thalweg = rm.plan.geom.rivers[nwm_id][nwm_id].ds_xs.thalweg
    df["ds_depth"] = df["ds_wse"] - thalweg

    insert_data(database_path, table_name, df, plan_suffix, missing_grids, boundary_condition="kwse")


def create_non_spatial_table(gpkg_path: str, metadata: dict) -> None:
    """Create the metadata table in the geopackage."""
    with sqlite3.connect(gpkg_path) as conn:
        string = ""
        curs = conn.cursor()
        curs.execute("DROP TABLE IF Exists metadata")
        curs.execute(f"CREATE TABLE IF NOT EXISTS metadata (key, value);")
        curs.close()

    with sqlite3.connect(gpkg_path) as conn:
        curs = conn.cursor()
        keys, vals = "", ""
        for key, val in metadata.items():

            if val:
                keys += f"{key},"
                vals += f"{val.replace(',','_')}, "
                curs.execute(f"INSERT INTO metadata (key,value) values (?,?);", (key, val))

        curs.execute("COMMIT;")
        curs.close()
    return None



"""Data model for shared utilities."""

import glob
import json
import math
import os
import sqlite3
from dataclasses import dataclass
from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import List

import geopandas as gpd
import pandas as pd
from pyproj import CRS
from shapely.geometry import LineString, Point

from ripple1d.utils.ripple_utils import (
    data_pairs_from_text_block,
    fix_reversed_xs,
    search_contents,
    text_block_from_start_end_str,
    text_block_from_start_str_length,
    text_block_from_start_str_to_empty_line,
    xs_concave_hull,
)
from ripple1d.utils.s3_utils import init_s3_resources, read_json_from_s3


class RasModelStructure:
    """Base Model structure for RAS models."""

    def __init__(self, model_directory: str):
        self.model_directory = model_directory
        self.model_basename = Path(model_directory).name
        self._ras_junctions = None
        self._ras_structures = None
        self._ras_xs = None
        self._ras_rivers = None
        self._xs_concave_hull = None

    @property
    def model_name(self):
        """Model name."""
        return self.model_basename.replace(".prj", "")

    def derive_path(self, extension: str):
        """Derive path."""
        return str(Path(self.model_directory) / f"{self.model_name}{extension}")

    def file_exists(self, file_path: str) -> bool:
        """Check if file exists."""
        if os.path.exists(file_path):
            return True
        return False

    @property
    def ras_project_file(self):
        """RAS Project file."""
        return self.derive_path(".prj")

    @property
    def ras_gpkg_file(self):
        """RAS GeoPackage file."""
        return self.derive_path(".gpkg")

    @property
    def ras_xs(self):
        """RAS XS Geodataframe."""
        if self._ras_xs is None:
            self._ras_xs = gpd.read_file(self.ras_gpkg_file, layer="XS")
        return self._ras_xs

    @property
    def ras_junctions(self):
        """RAS Junctions Geodataframe."""
        if "Junction" in fiona.listlayers(self.ras_gpkg_file):
            if self._ras_junctions is None:
                self._ras_junctions = gpd.read_file(self.ras_gpkg_file, layer="Junction")
            return self._ras_junctions

    @property
    def ras_structures(self):
        """RAS Structures Geodataframe."""
        if "Structure" in fiona.listlayers(self.ras_gpkg_file):
            if self._ras_structures is None:
                self._ras_structures = gpd.read_file(self.ras_gpkg_file, layer="Structure")
            return self._ras_structures

    @property
    def ras_rivers(self):
        """RAS Rivers Geodataframe."""
        if self._ras_rivers is None:
            self._ras_rivers = gpd.read_file(self.ras_gpkg_file, layer="River")
        return self._ras_rivers

    @property
    def xs_concave_hull(self):
        """XS Concave Hull."""
        if self._xs_concave_hull is None:
            self._xs_concave_hull = xs_concave_hull(fix_reversed_xs(self.ras_xs, self.ras_rivers))
        return self._xs_concave_hull

    @property
    def assets(self):
        """Model assets."""
        return glob.glob(f"{self.model_directory}/Terrain/*") + [
            f for f in glob.glob(f"{self.model_directory}/*") if not os.path.isdir(f)
        ]

    @property
    def terrain_assets(self):
        """Terrain assets."""
        return glob.glob(f"{self.model_directory}/Terrain/*")

    @property
    def thumbnail_png(self):
        """Thumbnail PNG."""
        return self.derive_path(".png")


class RippleSourceModel:
    """Source Model structure for Ripple to create NwmReachModel's."""

    def __init__(self, ras_project_file: str, crs: CRS):

        self.crs = crs
        self.ras_project_file = ras_project_file
        self.model_directory = Path(ras_project_file).parent.as_posix()
        self.model_basename = Path(ras_project_file).as_posix()

    @property
    def model_name(self):
        """Model name."""
        return self.model_basename.replace(".prj", "")

    def derive_path(self, extension: str):
        """Derive path."""
        return str(Path(self.model_directory) / f"{self.model_name}{extension}")

    def file_exists(self, file_path: str) -> bool:
        """Check if file exists."""
        if os.path.exists(file_path):
            return True
        return False

    @property
    def ras_gpkg_file(self):
        """RAS GeoPackage file."""
        return self.derive_path(".gpkg")

    @property
    def assets(self):
        """Model assets."""
        return glob.glob(f"{self.model_directory}/Terrain/*") + [
            f for f in glob.glob(f"{self.model_directory}/*") if not os.path.isdir(f)
        ]

    @property
    def terrain_assets(self):
        """Terrain assets."""
        return glob.glob(f"{self.model_directory}/Terrain/*")

    @property
    def thumbnail_png(self):
        """Thumbnail PNG."""
        return self.derive_path(".png")

    @property
    def conflation_file(self):
        """Conflation file."""
        return self.derive_path(".conflation.json")

    def nwm_conflation_parameters(self, nwm_id: str):
        """NWM Conflation parameters."""
        with open(self.conflation_file, "r") as f:
            conflation_parameters = json.loads(f.read())
        return conflation_parameters[nwm_id]


class RippleSourceDirectory:
    """Source Directory for Ripple to create NwmReachModel's. Should contain the conflation.json file and gpkg file for the source model."""

    def __init__(self, source_directory: str):

        self.source_directory = source_directory
        self.model_basename = os.path.basename(self.source_directory)

    @property
    def model_name(self):
        """Model name."""
        return self.model_basename

    def derive_path(self, extension: str):
        """Derive path."""
        return str(Path(self.source_directory) / f"{self.model_name}{extension}")

    def file_exists(self, file_path: str) -> bool:
        """Check if file exists."""
        if os.path.exists(file_path):
            return True
        return False

    @property
    def ras_gpkg_file(self):
        """RAS GeoPackage file."""
        return self.derive_path(".gpkg")

    @property
    def ras_project_file(self):
        """RAS Project file."""
        return self.derive_path(".prj")

    @property
    def assets(self):
        """Model assets."""
        return glob.glob(f"{self.source_directory}/Terrain/*") + [
            f for f in glob.glob(f"{self.source_directory}/*") if not os.path.isdir(f)
        ]

    @property
    def terrain_assets(self):
        """Terrain assets."""
        return glob.glob(f"{self.source_directory}/Terrain/*")

    @property
    def thumbnail_png(self):
        """Thumbnail PNG."""
        return self.derive_path(".png")

    @property
    def conflation_file(self):
        """Conflation file."""
        return self.derive_path(".conflation.json")

    def nwm_conflation_parameters(self, nwm_id: str):
        """NWM Conflation parameters."""
        with open(self.conflation_file, "r") as f:
            conflation_parameters = json.loads(f.read())
        return conflation_parameters["reaches"][nwm_id]


class NwmReachModel(RasModelStructure):
    """National Water Model reach-based HEC-RAS Model files and directory structure."""

    def __init__(self, model_directory: str, library_directory: str = ""):
        super().__init__(model_directory)
        self.library_directory = library_directory

    @property
    def terrain_directory(self):
        """Terrain directory."""
        return str(Path(self.model_directory) / "Terrain")

    @property
    def ras_terrain_hdf(self):
        """RAS Terrain HDF file."""
        return str(Path(self.terrain_directory) / f"{self.model_name}.hdf")

    @property
    def fim_results_directory(self):
        """FIM results directory."""
        return str(Path(self.library_directory) / self.model_name)

    @property
    def fim_lib_assets(self):
        """Assets of the fim library."""
        return glob.glob(f"{self.fim_results_directory}/*/*")

    @property
    def fim_lib_stac_json_file(self):
        """FIM LIBRARY STAC JSON file."""
        return str(Path(self.fim_results_directory) / f"{self.model_name}.fim_lib.stac.json")

    @property
    def fim_results_database(self):
        """Results database."""
        return str(Path(self.library_directory) / f"{self.model_name}.db")

    @property
    def fim_rating_curve(self):
        """FIM rating curve."""
        with sqlite3.connect(self.fim_results_database) as conn:
            cursor = conn.cursor()
            sql_query = f"""SELECT us_flow, us_wse, ds_wse
            FROM rating_curves
            WHERE reach_id={self.model_name}"""
            cursor.execute(sql_query)

            data = cursor.fetchall()
        data_list = ["Flow | US WSE | DS WSE"]
        for row in data:
            data_list.append(" | ".join([str(r) for r in row]))

        return data_list

    @property
    def crs(self):
        """Coordinate Reference System."""
        return self.ripple1d_parameters["crs"]

    def upload_files_to_s3(self, ras_s3_prefix: str, bucket: str):
        """Upload the model to s3."""
        _, client, _ = init_s3_resources()
        for file in self.assets:
            key = str(PurePosixPath(Path(file.replace(self.model_directory, ras_s3_prefix))))
            client.upload_file(
                Bucket=bucket,
                Key=key,
                Filename=file,
            )

    def upload_fim_lib_assets(self, s3_prefix: str, bucket: str):
        """Upload the fim lib to s3."""
        _, client, _ = init_s3_resources()
        for file in self.fim_lib_assets:
            parts = list(Path(file).parts[-2:])
            file_name = "-".join(parts)
            key = f"{s3_prefix}/{file_name}"

            client.upload_file(
                Bucket=bucket,
                Key=key,
                Filename=file,
            )
        file_name = os.path.basename(self.fim_lib_stac_json_file)
        client.upload_file(Bucket=bucket, Key=f"{s3_prefix}/{file_name}", Filename=self.fim_lib_stac_json_file)

    @property
    def ripple1d_parameters(self):
        """Ripple parameters."""
        with open(self.conflation_file, "r") as f:
            ripple1d_parameters = json.loads(f.read())
        return ripple1d_parameters

    def update_write_ripple1d_parameters(self, new_parameters: dict):
        """Write Ripple parameters."""
        parameters = self.ripple1d_parameters
        parameters.update(new_parameters)
        with open(self.conflation_file, "w") as f:
            f.write(json.dumps(parameters, indent=4))

    @property
    def conflation_file(self):
        """Conflation file."""
        return self.derive_path(".ripple1d.json")

    @property
    def model_stac_json_file(self):
        """STAC JSON file."""
        return self.derive_path(".model.stac.json")


@dataclass
class FlowChangeLocation:
    """HEC-RAS Flow Change Locations."""

    river: str = None
    reach: str = None
    rs: float = None
    flows: list[float] = None
    profile_names: list[str] = None


class XS:
    """HEC-RAS Cross Section."""

    def __init__(self, ras_data: list, river_reach: str, river: str, reach: str, crs: str):
        self.ras_data = ras_data
        self.crs = crs
        self.river = river
        self.reach = reach
        self.river_reach = river_reach
        self.river_reach_rs = f"{river} {reach} {self.river_station}"

    def split_xs_header(self, position: int):
        """
        Split cross section header.

        Example: Type RM Length L Ch R = 1 ,83554.  ,237.02,192.39,113.07.
        """
        header = search_contents(self.ras_data, "Type RM Length L Ch R ", expect_one=True)

        return header.split(",")[position]

    @property
    def river_station(self):
        """Cross section river station."""
        return float(self.split_xs_header(1).replace("*", ""))

    @property
    def left_reach_length(self):
        """Cross section left reach length."""
        return float(self.split_xs_header(2))

    @property
    def channel_reach_length(self):
        """Cross section channel reach length."""
        return float(self.split_xs_header(3))

    @property
    def right_reach_length(self):
        """Cross section right reach length."""
        return float(self.split_xs_header(4))

    @property
    def number_of_coords(self):
        """Number of coordinates in cross section."""
        try:
            return int(search_contents(self.ras_data, "XS GIS Cut Line", expect_one=True))
        except ValueError as e:
            return 0
            # raise NotGeoreferencedError(f"No coordinates found for cross section: {self.river_reach_rs} ")

    @property
    def thalweg(self):
        """Cross section thalweg elevation."""
        if self.station_elevation_points:
            _, y = list(zip(*self.station_elevation_points))
            return min(y)

    @property
    def xs_max_elevation(self):
        """Cross section maximum elevation."""
        if self.station_elevation_points:
            _, y = list(zip(*self.station_elevation_points))
            return max(y)

    @property
    def coords(self):
        """Cross section coordinates."""
        lines = text_block_from_start_str_length(
            f"XS GIS Cut Line={self.number_of_coords}",
            math.ceil(self.number_of_coords / 2),
            self.ras_data,
        )
        if lines:
            return data_pairs_from_text_block(lines, 32)

    @property
    def number_of_station_elevation_points(self):
        """Number of station elevation points."""
        return int(search_contents(self.ras_data, "#Sta/Elev", expect_one=True))

    @property
    def station_elevation_points(self):
        """Station elevation points."""
        try:
            lines = text_block_from_start_str_length(
                f"#Sta/Elev= {self.number_of_station_elevation_points} ",
                math.ceil(self.number_of_station_elevation_points / 5),
                self.ras_data,
            )
            return data_pairs_from_text_block(lines, 16)
        except ValueError as e:
            return None

    @property
    def bank_stations(self):
        """Bank stations."""
        return search_contents(self.ras_data, "Bank Sta", expect_one=True).split(",")

    @property
    def gdf(self):
        """Cross section geodataframe."""
        return gpd.GeoDataFrame(
            {
                "geometry": [LineString(self.coords)],
                "river": [self.river],
                "reach": [self.reach],
                "river_reach": [self.river_reach],
                "river_station": [self.river_station],
                "river_reach_rs": [self.river_reach_rs],
                "thalweg": [self.thalweg],
                "xs_max_elevation": [self.xs_max_elevation],
                "left_reach_length": [self.left_reach_length],
                "right_reach_length": [self.right_reach_length],
                "channel_reach_length": [self.channel_reach_length],
                "ras_data": ["\n".join(self.ras_data)],
                "station_elevation_points": [self.station_elevation_points],
                "bank_stations": [self.bank_stations],
                "number_of_station_elevation_points": [self.number_of_station_elevation_points],
                "number_of_coords": [self.number_of_coords],
                # "coords": [self.coords],
            },
            crs=self.crs,
            geometry="geometry",
        )


class Structure:
    """Structure."""

    def __init__(self, ras_data: list, river_reach: str, river: str, reach: str, crs: str, us_xs: XS):
        self.ras_data = ras_data
        self.crs = crs
        self.river = river
        self.reach = reach
        self.river_reach = river_reach
        self.river_reach_rs = f"{river} {reach} {self.river_station}"
        self.us_xs = us_xs

    def split_structure_header(self, position: int):
        """
        Split Structure header.

        Example: Type RM Length L Ch R = 3 ,83554.  ,237.02,192.39,113.07.
        """
        header = search_contents(self.ras_data, "Type RM Length L Ch R ", expect_one=True)

        return header.split(",")[position]

    @property
    def river_station(self):
        """Structure river station."""
        return float(self.split_structure_header(1))

    @property
    def type(self):
        """Structure type."""
        return int(self.split_structure_header(0))

    def structure_data(self, position: int):
        """Structure data."""
        if self.type in [2, 3, 4]:  # culvert or bridge
            data = text_block_from_start_str_length(
                "Deck Dist Width WeirC Skew NumUp NumDn MinLoCord MaxHiCord MaxSubmerge Is_Ogee", 1, self.ras_data
            )
            return data[0].split(",")[position]
        elif self.type == 5:  # inline weir
            data = text_block_from_start_str_length(
                "IW Dist,WD,Coef,Skew,MaxSub,Min_El,Is_Ogee,SpillHt,DesHd", 1, self.ras_data
            )
            return data[0].split(",")[position]
        elif self.type == 6:  # lateral structure
            return 0

    @property
    def distance(self):
        """Distance to upstream cross section."""
        return float(self.structure_data(0))

    @property
    def width(self):
        """Structure width."""
        # TODO check units of the RAS model
        return float(self.structure_data(1))

    @property
    def gdf(self):
        """Structure geodataframe."""
        return gpd.GeoDataFrame(
            {
                "geometry": [LineString(self.us_xs.coords).offset_curve(self.distance)],
                "river": [self.river],
                "reach": [self.reach],
                "river_reach": [self.river_reach],
                "river_station": [self.river_station],
                "river_reach_rs": [self.river_reach_rs],
                "type": [self.type],
                "distance": [self.distance],
                "width": [self.width],
                "ras_data": ["\n".join(self.ras_data)],
            },
            crs=self.crs,
            geometry="geometry",
        )


class Reach:
    """HEC-RAS River Reach."""

    def __init__(self, ras_data: list, river_reach: str, crs: str):
        reach_lines = text_block_from_start_end_str(f"River Reach={river_reach}", ["River Reach"], ras_data, -1)
        self.ras_data = reach_lines
        self.crs = crs
        self.river_reach = river_reach
        self.river = river_reach.split(",")[0].rstrip()
        self.reach = river_reach.split(",")[1].rstrip()

        us_connection: str = None
        ds_connection: str = None

    @property
    def us_xs(self):
        """Upstream cross section."""
        return self.cross_sections[
            self.xs_gdf.loc[
                self.xs_gdf["river_station"] == self.xs_gdf["river_station"].max(),
                "river_reach_rs",
            ][0]
        ]

    @property
    def ds_xs(self):
        """Downstream cross section."""
        return self.cross_sections[
            self.xs_gdf.loc[
                self.xs_gdf["river_station"] == self.xs_gdf["river_station"].min(),
                "river_reach_rs",
            ][0]
        ]

    @property
    def number_of_cross_sections(self):
        """Number of cross sections."""
        return len(self.cross_sections)

    @property
    def number_of_coords(self):
        """Number of coordinates in reach."""
        return int(search_contents(self.ras_data, "Reach XY"))

    @property
    def coords(self):
        """Reach coordinates."""
        lines = text_block_from_start_str_length(
            f"Reach XY= {self.number_of_coords} ",
            math.ceil(self.number_of_coords / 2),
            self.ras_data,
        )
        return data_pairs_from_text_block(lines, 32)

    @property
    def reach_nodes(self):
        """Reach nodes."""
        return search_contents(self.ras_data, "Type RM Length L Ch R ", expect_one=False)

    @property
    def cross_sections(self):
        """Cross sections."""
        cross_sections = {}
        for header in self.reach_nodes:
            type, _, _, _, _ = header.split(",")[:5]
            if int(type) != 1:
                continue
            xs_lines = text_block_from_start_end_str(
                f"Type RM Length L Ch R ={header}",
                ["Type RM Length L Ch R", "River Reach"],
                self.ras_data,
            )
            cross_section = XS(xs_lines, self.river_reach, self.river, self.reach, self.crs)
            cross_sections[cross_section.river_reach_rs] = cross_section

        return cross_sections

    @property
    def structures(self):
        """Structures."""
        structures = {}
        for header in self.reach_nodes:
            type, _, _, _, _ = header.split(",")[:5]
            if int(type) == 1:
                xs_lines = text_block_from_start_end_str(
                    f"Type RM Length L Ch R ={header}",
                    ["Type RM Length L Ch R", "River Reach"],
                    self.ras_data,
                )
                cross_section = XS(xs_lines, self.river_reach, self.river, self.reach, self.crs)
                continue
            elif int(type) in [2, 3, 4, 5, 6]:  # culvert or bridge or multiple openeing
                structure_lines = text_block_from_start_end_str(
                    f"Type RM Length L Ch R ={header}",
                    ["Type RM Length L Ch R", "River Reach"],
                    self.ras_data,
                )
            else:
                raise TypeError(
                    f"Unsupported structure type: {int(type)}. Supported structure types are 2, 3, 4, 5, and 6 corresponding to culvert, bridge, multiple openeing, inline structure, lateral structure, respectively"
                )

            structure = Structure(structure_lines, self.river_reach, self.river, self.reach, self.crs, cross_section)
            structures[structure.river_reach_rs] = structure

        return structures

    @property
    def gdf(self):
        """Reach geodataframe."""
        return gpd.GeoDataFrame(
            {
                "geometry": [LineString(self.coords)],
                "river": [self.river],
                "reach": [self.reach],
                "river_reach": [self.river_reach],
                # "number_of_coords": [self.number_of_coords],
                # "coords": [self.coords],
                "ras_data": ["\n".join(self.ras_data)],
            },
            crs=self.crs,
            geometry="geometry",
        )

    @property
    def xs_gdf(self):
        """Cross section geodataframe."""
        return pd.concat([xs.gdf for xs in self.cross_sections.values()])

    @property
    def structures_gdf(self):
        """Structures geodataframe."""
        return pd.concat([structure.gdf for structure in self.structures.values()])


class Junction:
    """HEC-RAS Junction."""

    def __init__(self, ras_data: List[str], junct: str, crs: str):
        self.crs = crs
        self.name = junct
        self.ras_data = text_block_from_start_str_to_empty_line(f"Junct Name={junct}", ras_data)

    def split_lines(self, lines: str, token: str, idx: int):
        """Split lines."""
        return list(map(lambda line: line.split(token)[idx].rstrip(), lines))

    @property
    def x(self):
        """Junction x coordinate."""
        return self.split_lines([search_contents(self.ras_data, "Junct X Y & Text X Y")], ",", 0)

    @property
    def y(self):
        """Junction y coordinate."""
        return self.split_lines([search_contents(self.ras_data, "Junct X Y & Text X Y")], ",", 1)

    @property
    def point(self):
        """Junction point."""
        return Point(self.x, self.y)

    @property
    def upstream_rivers(self):
        """Upstream rivers."""
        return ",".join(
            self.split_lines(
                search_contents(self.ras_data, "Up River,Reach", expect_one=False),
                ",",
                0,
            )
        )

    @property
    def downstream_rivers(self):
        """Downstream rivers."""
        return ",".join(
            self.split_lines(
                search_contents(self.ras_data, "Dn River,Reach", expect_one=False),
                ",",
                0,
            )
        )

    @property
    def upstream_reaches(self):
        """Upstream reaches."""
        return ",".join(
            self.split_lines(
                search_contents(self.ras_data, "Up River,Reach", expect_one=False),
                ",",
                1,
            )
        )

    @property
    def downstream_reaches(self):
        """Downstream reaches."""
        return ",".join(
            self.split_lines(
                search_contents(self.ras_data, "Dn River,Reach", expect_one=False),
                ",",
                1,
            )
        )

    @property
    def junction_lengths(self):
        """Junction lengths."""
        return ",".join(self.split_lines(search_contents(self.ras_data, "Junc L&A", expect_one=False), ",", 0))

    @property
    def gdf(self):
        """Junction geodataframe."""
        return gpd.GeoDataFrame(
            {
                "geometry": [self.point],
                "junction_lengths": [self.junction_lengths],
                "us_rivers": [self.upstream_rivers],
                "ds_rivers": [self.downstream_rivers],
                "us_reaches": [self.upstream_reaches],
                "ds_reaches": [self.downstream_reaches],
                "ras_data": ["\n".join(self.ras_data)],
            },
            geometry="geometry",
            crs=self.crs,
        )



"""Create FIM library."""

import glob
import json
import logging
import os
import shutil
from datetime import datetime
from pathlib import Path, PurePosixPath

import geopandas as gpd
import pystac
import rasterio
import shapely
from pyproj import CRS

# from osgeo import gdal
from rasterio.enums import Resampling
from rasterio.shutil import copy as copy_raster

import ripple1d
from ripple1d.conflate.rasfim import RasFimConflater
from ripple1d.data_model import NwmReachModel
from ripple1d.errors import DepthGridNotFoundError, PlanNameNotFoundError
from ripple1d.ras import RasManager
from ripple1d.ras_to_gpkg import geom_flow_to_gdfs, new_stac_item
from ripple1d.utils.dg_utils import (
    bbox_to_polygon,
    get_raster_bounds,
    get_raster_metadata,
    reproject_raster,
)
from ripple1d.utils.s3_utils import get_basic_object_metadata, init_s3_resources
from ripple1d.utils.sqlite_utils import (
    create_db_and_table,
    rating_curves_to_sqlite,
    zero_depth_to_sqlite,
)


def post_process_depth_grids(
    rm: RasManager,
    plan_name: str,
    dest_directory: str,
    accept_missing_grid: bool = False,
    overviews=False,
    dest_crs: CRS = 5070,
    resolution: float = 3,
    resolution_units: str = "Meters",
) -> tuple[list[str]]:
    """Clip depth grids based on their associated NWM branch and respective cross sections."""
    if resolution and not resolution_units:
        raise ValueError(
            f"The 'resolution' arg has been provided but 'resolution_units' arg has not been provided. Please provide both"
        )

    if resolution_units:
        if resolution_units not in ["Feet", "Meters"]:
            raise ValueError(f"Invalid resolution_units: {resolution_units}. expected 'Feet' or 'Meters'")

    for profile_name in rm.plans[plan_name].flow.profile_names:
        # construct the default path to the depth grid for this plan/profile
        src_dir = os.path.join(rm.ras_project._ras_dir, str(plan_name))
        terrain_dir = os.path.join(rm.ras_project._ras_dir, "Terrain")
        terrain_part = os.path.basename(glob.glob(terrain_dir + "\\*.tif")[0]).split(".")[-2]
        src_path = os.path.join(
            src_dir,
            f"Depth ({profile_name}).{rm.ras_project.title}.{terrain_part}.tif",
        )

        # if the depth grid path does not exists print a warning then continue to the next profile
        if not os.path.exists(src_path):
            if accept_missing_grid:
                logging.warning(f"depth raster does not exists: {src_path}")
                continue
            else:
                raise DepthGridNotFoundError(f"depth raster does not exists: {src_path}")

        if "kwse" in plan_name:
            flow, depth = profile_name.split("-", 1)
        elif "nd" in plan_name:
            flow = f"f_{profile_name}"
            depth = "z_nd"

        flow_sub_directory = os.path.join(dest_directory, depth)
        os.makedirs(flow_sub_directory, exist_ok=True)
        dest_path = os.path.join(flow_sub_directory, f"{flow}.tif")
        logging.debug(dest_path)
        reproject_raster(src_path, dest_path, CRS(dest_crs), resolution, resolution_units)
        logging.debug(f"Building overviews for: {dest_path}")

        if overviews:
            with rasterio.open(dest_path, "r+") as dst:
                dst.build_overviews([4, 8, 16], Resampling.nearest)
                dst.update_tags(ns="rio_overview", resampling="nearest")


def create_rating_curves_db(
    submodel_directory: str, plans: list, ras_version: str = "631", table_name: str = "rating_curves"
):
    """Create a new rating curve database for a NWM id.

    Export stage-discharge rating curves from HEC-RAS to
    rasters and a sqlite database.

    Parameters
    ----------
    submodel_directory : str
        The path to the directory containing a sub model geopackage
    plans : list
        suffixes of plans to create fim library for.
    ras_version : str, optional
        which version of HEC-RAS to use, by default "631"
    table_name : str, optional
        name for the table holding stage-discharge rating curves in the output
        database, by default "rating_curves"

    Returns
    -------
    dict
        dictionary with paths to output rating curve database
    """
    logging.info(f"create_rating_curves_db starting")

    nwm_rm = NwmReachModel(submodel_directory, submodel_directory)

    rm = RasManager(
        nwm_rm.ras_project_file,
        version=ras_version,
        terrain_path=nwm_rm.ras_terrain_hdf,
        crs=nwm_rm.crs,
    )

    # create dabase and table
    if not os.path.exists(nwm_rm.fim_results_database):
        create_db_and_table(nwm_rm.fim_results_database, table_name)

    for plan in plans:
        if f"{nwm_rm.model_name}_{plan}" not in rm.plans:
            logging.error(f"Plan {nwm_rm.model_name}_{plan} not found in the model, skipping...")
            continue
        else:
            missing_grids = find_missing_grids(rm, f"{nwm_rm.model_name}_{plan}")

        if f"kwse" in plan:
            rating_curves_to_sqlite(
                rm,
                f"{nwm_rm.model_name}_{plan}",
                plan,
                nwm_rm.model_name,
                missing_grids,
                nwm_rm.fim_results_database,
                table_name,
            )
        if f"nd" in plan:
            zero_depth_to_sqlite(
                rm,
                f"{nwm_rm.model_name}_{plan}",
                plan,
                nwm_rm.model_name,
                missing_grids,
                nwm_rm.fim_results_database,
                table_name,
            )

    logging.info(f"create_rating_curves_db complete")
    return {"rating_curve_database": nwm_rm.fim_results_database}


def find_missing_grids(
    rm: RasManager,
    plan_name: str,
):
    """Find missing depth grids."""
    missing_grids = []
    for profile_name in rm.plans[plan_name].flow.profile_names:
        # construct the default path to the depth grid for this plan/profile
        src_dir = os.path.join(rm.ras_project._ras_dir, str(plan_name))
        terrain_dir = os.path.join(rm.ras_project._ras_dir, "Terrain")
        terrain_part = os.path.basename(glob.glob(terrain_dir + "\\*.tif")[0]).split(".")[-2]
        src_path = os.path.join(
            src_dir,
            f"Depth ({profile_name}).{rm.ras_project.title}.{terrain_part}.tif",
        )

        # if the depth grid path does not exists print a warning then continue to the next profile
        if not os.path.exists(src_path):
            missing_grids.append(profile_name)

    return missing_grids


def create_fim_lib(
    submodel_directory: str,
    plans: list,
    library_directory: str,
    cleanup: bool,
    ras_version: str = "631",
    overviews: bool = False,
    resolution: float = 3,
    resolution_units: str = "Meters",
    dest_crs: str = 5070,
):
    """Create a new FIM library for a NWM id.

    Export depth rasters and stage-discharge rating curves from HEC-RAS to
    rasters and a sqlite database.

    Parameters
    ----------
    submodel_directory : str
        The path to the directory containing a sub model geopackage
    plans : list
        suffixes of plans to create fim library for.
    library_directory : str
        No function
    cleanup : bool
        whether to delete the source depth grids once they've been processed
    ras_version : str, optional
        which version of HEC-RAS to use, by default "631"
    overviews : bool, optional
        whether to generate overviews for output rasters (overviews at levels
        [4, 8, 16]), by default False
    resolution : float, optional
        horizontal resolution to resample output raster to, by default 3
    resolution_units : str, optional
        unit for resolution, by default "Meters"
    dest_crs : str, optional
        Destination crs.

    Returns
    -------
    dict
        dictionary with paths to output rasters and rating curve database
    """
    logging.info(f"create_fim_lib starting")
    nwm_rm = NwmReachModel(submodel_directory, library_directory)

    rm = RasManager(
        nwm_rm.ras_project_file,
        version=ras_version,
        terrain_path=nwm_rm.ras_terrain_hdf,
        crs=nwm_rm.crs,
    )

    for plan in plans:
        if f"{nwm_rm.model_name}_{plan}" not in rm.plans:
            logging.error(f"Plan {nwm_rm.model_name}_{plan} not found in the model, skipping...")
            continue
        else:
            post_process_depth_grids(
                rm,
                f"{nwm_rm.model_name}_{plan}",
                nwm_rm.fim_results_directory,
                accept_missing_grid=True,
                overviews=overviews,
                resolution=resolution,
                resolution_units=resolution_units,
                dest_crs=dest_crs,
            )

        if cleanup:
            shutil.rmtree(os.path.join(rm.ras_project._ras_dir, f"{nwm_rm.model_name}_{plan}"), ignore_errors=True)

    logging.info(f"create_fim_lib complete")

    return {"fim_results_directory": nwm_rm.fim_results_directory}


def nwm_reach_model_stac(ras_project_directory: str, ras_model_s3_prefix: str = None, bucket: str = None):
    """Convert a FIM RAS model to a STAC item."""
    logging.info(f"nwm_reach_model_stac starting")
    nwm_rm = NwmReachModel(ras_project_directory)

    # create new stac item
    new_stac_item(
        ras_project_directory,
        ras_model_s3_prefix,
    )

    # upload to s3
    if bucket and ras_model_s3_prefix:
        nwm_rm.upload_files_to_s3(ras_model_s3_prefix, bucket)

        stac_item = pystac.read_file(nwm_rm.model_stac_json_file)
        # update asset hrefs
        for id, asset in stac_item.assets.items():
            if "thumbnail" in asset.roles:
                asset.href = f"https://{bucket}.s3.amazonaws.com/{ras_model_s3_prefix}/{Path(asset.href).name}"
            else:
                asset.href = f"s3://{bucket}/{ras_model_s3_prefix}/{Path(asset.href).name}"

        stac_item.set_self_href(
            f"https://{bucket}.s3.amazonaws.com/{ras_model_s3_prefix}/{Path(stac_item.self_href).name}"
        )
        # write updated stac item to s3
        _, s3_client, _ = init_s3_resources()
        key = f"{ras_model_s3_prefix}/{Path(nwm_rm.model_stac_json_file).name}"
        s3_client.put_object(
            Body=json.dumps(stac_item.to_dict()).encode(),
            Bucket=bucket,
            Key=key,
        )
        nwm_rm.update_write_ripple1d_parameters({"model_stac_item": f"https://{bucket}.s3.amazonaws.com/{key}"})
    else:
        nwm_rm.update_write_ripple1d_parameters({"model_stac_item": nwm_rm.model_stac_json_file})
    logging.info(f"nwm_reach_model_stac complete")


def update_stac_s3_location(stac_item_file: pystac.Item, bucket: str, s3_prefix: str):
    """Update the href locations for a stac item and its assets."""
    stac_item = pystac.read_file(stac_item_file)
    _, s3_client, s3_resource = init_s3_resources()
    s3_bucket = s3_resource.Bucket(bucket)

    # update asset hrefs
    for asset in stac_item.assets.values():
        parts = list(Path(asset.href).parts[-2:])
        file_name = "-".join(parts)
        s3_key = f"{s3_prefix}/{file_name}"

        s3_uri = f"s3://{bucket}/{s3_key}"
        if "thumbnail" in asset.roles:
            asset.href = f"https://{bucket}.s3.amazonaws.com/{s3_key}"
        else:
            asset.href = s3_uri

        s3_obj = s3_bucket.Object(s3_key)
        metadata = get_basic_object_metadata(s3_obj)
        asset.extra_fields.update(metadata)

    stac_item.set_self_href(f"https://{bucket}.s3.amazonaws.com/{s3_prefix}/{Path(stac_item.self_href).name}")
    # write updated stac item to s3
    s3_client.put_object(
        Body=json.dumps(stac_item.to_dict()).encode(),
        Bucket=bucket,
        Key=f"{s3_prefix}/{Path(stac_item_file).name}",
    )


def fim_lib_item(item_id: str, assets: list, stac_json: str, metadata: dict) -> pystac.Item:
    """
    Create a PySTAC Item for a fim_lib_item.

    Parameters
    ----------
        item_id (str): The ID to assign to the PySTAC Item.
        assets (str): assets (files) to add to the stac item
        stac_json (str): json file to write the stac item to

    Returns
    -------
        pystac.Item: The PySTAC Item representing the raster file. The Item has an Asset with the raster file,
        the title set to the name of the raster file, the media type set to COG, and the role
        set to "ras-depth-grid". The Asset's extra fields are updated with the basic object metadata of the raster file
        and the metadata of the raster file. The Item's bbox is set to the geographic bounds of the raster file in the
        WGS 84 (EPSG:4326) coordinate reference system, the datetime is set to the current datetime, and the geometry
        is set to the GeoJSON representation of the bbox.
    """
    bbox = get_raster_bounds(assets[0])
    geometry = bbox_to_polygon(bbox)
    item = pystac.Item(
        id=item_id,
        properties=metadata["properties"],
        bbox=bbox,
        datetime=datetime.now(),
        geometry=json.loads(shapely.to_geojson(geometry)),
    )
    # add assets
    for file in assets:
        parts = list(Path(file).parts[-2:])
        title = "-".join(parts).rstrip(".tif")
        asset = pystac.Asset(
            href=file,
            title=title,
            media_type=pystac.MediaType.COG,
            roles=["fim"],
        )

        asset_metadata = get_raster_metadata(file)
        if asset_metadata:
            asset.extra_fields.update(asset_metadata)
        item.add_asset(key=asset.title, asset=asset)
    item.add_derived_from(metadata["derived_from"]["model_stac_item"])

    with open(stac_json, "w") as f:
        f.write(json.dumps(item.to_dict()))
    return item


def fim_lib_stac(ras_project_directory: str, nwm_reach_id: str, s3_prefix: str = None, bucket: str = None):
    """Create a stac item for a fim library."""
    logging.info(f"fim_lib_stac starting")
    nwm_rm = NwmReachModel(ras_project_directory)

    metadata = {
        "properties": {
            "NOAA_NWM:FIM Reach ID": nwm_reach_id,
            "NOAA_NWM:FIM to Reach ID": nwm_rm.ripple1d_parameters["nwm_to_id"],
            "NOAA_NWM:FIM Depth units": "ft",
            "NOAA_NWM:FIM Flow units": "cfs",
            "NOAA_NWM:FIM Rating Curve (Flow, Depth)": nwm_rm.fim_rating_curve,
            "proj:wkt2": CRS(nwm_rm.crs).to_wkt(),
            "proj:epsg": CRS(nwm_rm.crs).to_epsg(),
            "Ripple Version": ripple1d.__version__,
        },
        "derived_from": {"model_stac_item": nwm_rm.ripple1d_parameters["model_stac_item"]},
    }
    assets = nwm_rm.fim_lib_assets
    fim_lib_item(nwm_reach_id, nwm_rm.fim_lib_assets, nwm_rm.fim_lib_stac_json_file, metadata)

    if s3_prefix and bucket:
        nwm_rm.upload_fim_lib_assets(s3_prefix, bucket)
        update_stac_s3_location(nwm_rm.fim_lib_stac_json_file, bucket, s3_prefix)
    logging.info("fim_lib_stac complete")



"""Conflation Metrics."""

import json
import logging
import os
import sqlite3
import traceback

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
from shapely import Geometry, LineString, MultiLineString, MultiPoint, Point
from shapely.ops import linemerge

from ripple1d.consts import HYDROFABRIC_CRS, METERS_PER_FOOT
from ripple1d.data_model import XS
from ripple1d.ops.subset_gpkg import RippleGeopackageSubsetter
from ripple1d.utils.ripple_utils import fix_reversed_xs, xs_concave_hull


class ConflationMetrics:
    """Calculate metrics for a cross section."""

    def __init__(
        self,
        xs_gdf: gpd.GeoDataFrame,
        river_gdf: gpd.GeoDataFrame,
        hull_gdf,
        network_reach: LineString,
        network_reach_plus_ds_reach: LineString,
        network_id: str,
    ):
        self.xs_gdf = xs_gdf
        self.river_gdf = river_gdf
        self.hull_gdf = hull_gdf
        self.network_reach = network_reach
        self.network_reach_plus_ds_reach = network_reach_plus_ds_reach
        self.crs = xs_gdf.crs
        self.network_id = network_id

    def populate_station_elevation(self, row: pd.Series) -> dict:
        """Populate the station elevation for a cross section."""
        xs = XS(row.ras_data.splitlines(), row.river_reach, row.river, row.reach, self.crs)
        return dict(xs.station_elevation_points)

    def populate_thalweg_station(self, row: pd.Series) -> str:
        """Populate the thalweg station for a cross section."""
        df = pd.DataFrame(row["station_elevation"], index=["elevation"]).T
        return df.loc[df["elevation"] == row.thalweg].index[0]

    def thalweg_metrics(self, xs_gdf: gpd.GeoDataFrame) -> dict:
        """Calculate the distance between the thalweg point and the network intersection point."""
        try:
            xs_gdf["station_elevation"] = xs_gdf.apply(lambda row: self.populate_station_elevation(row), axis=1)
            xs_gdf["thalweg_station"] = xs_gdf.apply(lambda row: self.populate_thalweg_station(row), axis=1)
            xs_gdf["thalweg_point"] = xs_gdf.apply(
                lambda row: row.geometry.interpolate(row["thalweg_station"] * METERS_PER_FOOT), axis=1
            )

            xs_gdf["ras_intersection_point"] = None
            for _, r in self.river_gdf.iterrows():
                xs_gdf.loc[xs_gdf["river_reach"] == r["river_reach"], "ras_intersection_point"] = xs_gdf.loc[
                    xs_gdf["river_reach"] == r["river_reach"], :
                ].apply(lambda row: r.geometry.intersection(row.geometry), axis=1)

            xs_gdf["network_intersection_point"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.intersection(row.geometry), axis=1
            )

            xs_gdf["centerline_offset"] = xs_gdf.apply(
                lambda row: row["ras_intersection_point"].distance(row["network_intersection_point"]), axis=1
            )
            xs_gdf["thalweg_offset"] = xs_gdf.apply(
                lambda row: row["thalweg_point"].distance(row["network_intersection_point"]), axis=1
            )

            return {
                "centerline_offset": xs_gdf["centerline_offset"].describe().round().fillna(-9999).astype(int).to_dict(),
                "thalweg_offset": xs_gdf["thalweg_offset"].describe().round().fillna(-9999).astype(int).to_dict(),
            }
        except Exception as e:
            logging.error(f"network id: {self.network_id} | Error: {e}")
            logging.error(f"network id: {self.network_id} | Traceback: {traceback.format_exc()}")

    def clean_length(self, length: float, meters_to_feet: bool = True):
        """Clean the length."""
        if length == 0:
            length = None
        elif np.isnan(length):
            length = None
        else:
            if meters_to_feet:
                length = int(length / METERS_PER_FOOT)
            else:
                length = int(length)
        return length

    def compute_station(self, river: LineString, xs: gpd.GeoDataFrame):
        """Compute the station for a river reach."""
        xs["intersection_point"] = xs.apply(lambda row: river.intersection(row.geometry), axis=1)

        xs["station"] = xs.apply(
            lambda row: (
                river.project(row["intersection_point"]) if isinstance(row["intersection_point"], Point) else np.nan
            ),
            axis=1,
        )
        return self.clean_length(xs["station"].max() - xs["station"].min())

    @property
    def river_line(self):
        """Get the geomtry for the river centerline."""
        return self.river_gdf.iloc[0].geometry

    def us_ds_xs(self, xs_gdf: gpd.GeoDataFrame):
        """Get the most upstream and downstream cross sections."""
        if len(xs_gdf["river_reach"].unique()) > 1:
            raise ValueError("Cross sections must all be on the same river reach.")

        return xs_gdf[
            (xs_gdf["river_station"] == xs_gdf["river_station"].max())
            | (xs_gdf["river_station"] == xs_gdf["river_station"].min())
        ]

    def length_metrics(self, xs_gdf: gpd.GeoDataFrame) -> dict:
        """Calculate the reach length between cross sections along the ras river line and the network reach."""
        try:
            xs = self.us_ds_xs(xs_gdf)
            network_length = self.compute_station(self.network_reach_plus_ds_reach, xs)
            ras_length = self.compute_station(self.river_line, xs)

            if ras_length is None or network_length is None:
                network_ras_ratio = None
            else:
                network_ras_ratio = round(float(network_length / ras_length), 2)

            return {
                "ras": ras_length,
                "network": network_length,
                "network_to_ras_ratio": network_ras_ratio,
            }
        except Exception as e:
            logging.error(f"network id: {self.network_id} | Error: {e}")
            logging.error(f"network id: {self.network_id} | Traceback: {traceback.format_exc()}")

    # def parrallel_reaches(self, network_reaches: gpd.GeoDataFrame) -> dict:
    #     """Calculate the overlap between the network reach and the cross sections."""
    #     if network_reaches.empty:
    #         return []
    #     overlaps = network_reaches[network_reaches.intersects(self.xs_gdf.union_all())]
    #     if overlaps.empty:
    #         return []
    #     overlaps["overlap"] = overlaps.apply(
    #         lambda row: int(
    #             row["geometry"].intersection(xs_concave_hull(self.xs_gdf)["geometry"].iloc[0]).length / METERS_PER_FOOT
    #         ),
    #         axis=1,
    #     )

    #     return [{"id": str(row["ID"]), "overlap": row["overlap"]} for _, row in overlaps.iterrows()]

    def overlapped_reaches(self, to_reaches: gpd.GeoDataFrame) -> dict:
        """Calculate the overlap between the network reach and the cross sections."""
        if to_reaches.empty:
            return []
        reach_ids = []
        geom_name = to_reaches.geometry.name
        for i, row in to_reaches.iterrows():
            if row[geom_name].intersects(self.xs_gdf.union_all()):
                overlap = row[geom_name].intersection(self.hull_gdf["geometry"].iloc[0]).length / METERS_PER_FOOT
                reach_ids.append({"id": str(row["ID"]), "overlap": int(overlap)})
        return reach_ids

    def eclipsed_reaches(self, network_reaches: gpd.GeoDataFrame) -> dict:
        """Calculate the overlap between the network reach and the cross sections."""
        if network_reaches.empty:
            return []
        eclipsed_reaches = network_reaches[network_reaches.covered_by(self.hull_gdf["geometry"].iloc[0])]

        return [str(row["ID"]) for _, row in eclipsed_reaches.iterrows()]

    def compute_coverage_metrics(self, xs_gdf: gpd.GeoDataFrame) -> dict:
        """Calculate the coverage metrics for a set of cross sections."""
        try:
            xs_gdf = xs_gdf[
                (xs_gdf["river_station"] == xs_gdf["river_station"].max())
                | (xs_gdf["river_station"] == xs_gdf["river_station"].min())
            ]
            xs_gdf["intersection_point"] = xs_gdf.apply(
                lambda row: self.network_reach_plus_ds_reach.intersection(row.geometry), axis=1
            )
            xs_gdf["station_percent"] = xs_gdf.apply(
                lambda row: (
                    self.network_reach_plus_ds_reach.project(row["intersection_point"]) / self.network_reach.length
                    if isinstance(row["intersection_point"], Point)
                    else np.nan
                ),
                axis=1,
            )
            return {
                "start": float(round(xs_gdf["station_percent"].min(), 2)),
                "end": min([round(xs_gdf["station_percent"].max(), 2), 1]),
            }
        except Exception as e:
            logging.error(f"network id: {self.network_id} | Error: {e}")
            logging.error(f"network id: {self.network_id} | Traceback: {traceback.format_exc()}")


def compute_conflation_metrics(source_model_directory: str, source_network: dict):
    """Compute metrics for a network reach.

    Parameters
    ----------
    source_model_directory : str
        The path to the directory containing HEC-RAS project, plan, geometry,
        and flow files.
    source_network : dict
        Information on the network to conflate

        - **file_name** (str):
            path/to/nwm_network.parquet (required)
        - **type** (str):
            must be 'nwm_hydrofabric' (required)
        - **version** (str):
            optional version number to log
    task_id : str, optional
        Task ID to use for logging, by default ""

    Returns
    -------
    dict
        Dictionary containing metrics describing model network match

    Notes
    -----
    The compute_conflation_metrics endpoint uses the information gathered in
    conflation to summarize how well a NWM reach and HEC-RAS model section
    align.  The metrics are broken down into three categories: xs, lengths, and
    coverage.

    * **xs.**  These metrics quantify the degree of alignment between the NWM
      reach centerline and the HEC-RAS model.  The metrics below are measured
      at HEC-RAS cross-section and summary statistics are reported in the
      conflation metrics output.

      * centerline_offset measures the straightline distance between RAS
        centerline and NWM reach line
      * thalweg_offset measures the straightline distance between lowest point
        along each RAS section and NWM reach line
    * **lengths.** These metrics assess centerline length differences between
      HEC-RAS and the NWM reaches.

      * ras is the distance along the RAS centerline between upstream and
        downstream cross-section
      * network is the distance along the NWM reach between upstream and
        downstream cross-section
      * network_to_ras_ratio is the network length divided by ras length
    * **coverage.** These metrics quantify the portion of the NWM reach between
      the upstream and downstream cross-section.

      * start is the ratio of NWM reach length that occurs u/s of the upstream
        cross-section
      * end is the ratio of NWM reach length that occurs u/s of the downstream
        cross-section

    compute_conflation_metrics also reports overlapped reaches and eclipsed
    reaches.  Overlapped_reaches is a list of reaches that are intersected by
    the most d/s cross-section of this model.  Eclipsed_reaches is a list of
    NWM reaches that are contained within the concave hull of the
    cross-sections.
    """
    logging.info(f"compute_conflation_metrics starting")
    network_pq_path = source_network["file_name"]
    model_name = os.path.basename(source_model_directory)
    src_gpkg_path = os.path.join(source_model_directory, f"{model_name}.gpkg")
    conflation_json = os.path.join(source_model_directory, f"{model_name}.conflation.json")
    conflation_parameters = json.load(open(conflation_json))
    rgs = RippleGeopackageSubsetter(src_gpkg_path, conflation_json, "")

    for network_id in conflation_parameters["reaches"].keys():
        try:
            if conflation_parameters["reaches"][network_id]["eclipsed"] == True:
                continue

            rgs.set_nwm_id(network_id)
            layers = {}
            for layer, gdf in rgs.subset_gdfs.items():
                layers[layer] = gdf.to_crs(HYDROFABRIC_CRS)

            network_reaches = gpd.read_parquet(network_pq_path, bbox=layers["XS"].total_bounds)
            network_reach = linemerge(network_reaches.loc[network_reaches["ID"] == int(network_id)].geometry.iloc[0])
            network_reach_plus_ds_reach = combine_reaches(network_reaches, network_id)

            cm = ConflationMetrics(
                fix_reversed_xs(layers["XS"], layers["River"]),
                layers["River"],
                rgs.ripple_xs_concave_hull.to_crs(HYDROFABRIC_CRS),
                network_reach,
                network_reach_plus_ds_reach,
                network_id,
            )

            metrics = {
                "xs": cm.thalweg_metrics(layers["XS"]),
                "lengths": cm.length_metrics(layers["XS"]),
                "coverage": cm.compute_coverage_metrics(layers["XS"]),
            }

            to_id = conflation_parameters["reaches"][network_id]["network_to_id"]
            if to_id in conflation_parameters["reaches"].keys():
                next_to_id = int(conflation_parameters["reaches"][to_id]["network_to_id"])
            else:
                next_to_id = None

            overlapped_reaches = cm.overlapped_reaches(
                network_reaches[network_reaches["ID"].isin([int(to_id), next_to_id])]
            )
            eclipsed_reaches = get_eclipsed_reaches(conflation_parameters, network_id)

            conflation_parameters["reaches"][network_id].update({"metrics": metrics})
            conflation_parameters["reaches"][network_id].update({"overlapped_reaches": overlapped_reaches})
            conflation_parameters["reaches"][network_id].update({"eclipsed_reaches": eclipsed_reaches})
            conflation_parameters["metadata"]["length_units"] = "feet"
            conflation_parameters["metadata"]["flow_units"] = "cfs"
        except Exception as e:
            logging.error(f"network id: {network_id} | Error: {e}")
            logging.error(f"network id: {network_id} | Traceback: {traceback.format_exc()}")
            conflation_parameters["reaches"][network_id].update({"metrics": {}})
    with open(conflation_json, "w") as f:
        f.write(json.dumps(conflation_parameters, indent=4))

    logging.info(f"compute_conflation_metrics complete")
    return conflation_parameters


def combine_reaches(network_reaches: gpd.GeoDataFrame, network_id: str) -> LineString:
    """Combine network reaches."""
    reach = network_reaches.loc[network_reaches["ID"] == int(network_id), :]
    to_reach = network_reaches.loc[network_reaches["ID"] == int(reach["to_id"].iloc[0]), :]

    if to_reach.empty:
        return linemerge(reach.geometry.iloc[0])
    else:
        next_to_reach = network_reaches.loc[network_reaches["ID"] == int(to_reach["to_id"].iloc[0]), :]
        if next_to_reach.empty:
            return linemerge(MultiLineString([linemerge(reach.geometry.iloc[0]), linemerge(to_reach.geometry.iloc[0])]))
        else:
            return linemerge(
                MultiLineString(
                    [
                        linemerge(reach.geometry.iloc[0]),
                        linemerge(to_reach.geometry.iloc[0]),
                        linemerge(next_to_reach.geometry.iloc[0]),
                    ]
                )
            )


def get_eclipsed_reaches(conflation_parameters, network_id):
    """Walk the network to find all eclipsed reaches until the next non-eclipsed reach."""
    walking = True
    next_ds = conflation_parameters["reaches"][network_id]["network_to_id"]
    eclipsed = []
    while walking:
        if not next_ds in conflation_parameters["reaches"]:
            break
        if not conflation_parameters["reaches"][next_ds]["eclipsed"]:
            break
        else:
            eclipsed.append(next_ds)
            next_ds = conflation_parameters["reaches"][next_ds]["network_to_id"]
    return eclipsed
